from enum import IntEnum

from bidict import bidict


class Unit(IntEnum):
    MOVEABLE_MAP_REVEALER = 0
    ALFRED_THE_ALPACA = 1300
    AMAZON_ARCHER = 850
    AMAZON_WARRIOR = 825
    ARAMBAI = 1126
    ARBALESTER = 492
    ARCHER = 4
    BACTRIAN_CAMEL = 1237
    BALLISTA_ELEPHANT = 1120
    BANDIT = 299
    BATTERING_RAM = 1258
    BATTLE_ELEPHANT = 1132
    BERSERK = 692
    BOMBARD_CANNON = 36
    BOYAR = 876
    BUILDER = 118
    CAMEL = 897
    CAMEL_ARCHER = 1007
    CAMEL_RIDER = 329
    CANNON_GALLEON = 420
    CANOE = 778
    CAPPED_RAM = 422
    CARAVEL = 1004
    CART = 1338
    CATAPHRACT = 40
    CAVALIER = 283
    CAVALRY_ARCHER = 39
    CENTURION = 275
    CHAMPION = 567
    CHU_KO_NU = 73
    COBRA_CAR = 748
    CONDOTTIERO = 882
    CONQUISTADOR = 771
    COW_A = 705
    COW_B = 1596
    COW_C = 1598
    COW_D = 1600
    CROSSBOWMAN = 24
    DEMOLITION_RAFT = 1104
    DEMOLITION_SHIP = 527
    DONKEY = 846
    DRAGON_SHIP = 1302
    EAGLE_SCOUT = 751
    EAGLE_WARRIOR = 753
    EASTERN_SWORDSMAN = 894
    ELEPHANT_ARCHER = 873
    ELITE_ARAMBAI = 1128
    ELITE_BALLISTA_ELEPHANT = 1122
    ELITE_BATTLE_ELEPHANT = 1134
    ELITE_BERSERK = 694
    ELITE_BOYAR = 878
    ELITE_CAMEL_ARCHER = 1009
    ELITE_CANNON_GALLEON = 691
    ELITE_CARAVEL = 1006
    ELITE_CATAPHRACT = 553
    ELITE_CHU_KO_NU = 559
    ELITE_CONQUISTADOR = 773
    ELITE_EAGLE_WARRIOR = 752
    ELITE_ELEPHANT_ARCHER = 875
    ELITE_GBETO = 1015
    ELITE_GENITOUR = 1012
    ELITE_GENOESE_CROSSBOWMAN = 868
    ELITE_HUSKARL = 555
    ELITE_JAGUAR_WARRIOR = 726
    ELITE_JANISSARY = 557
    ELITE_KAMAYUK = 881
    ELITE_KARAMBIT_WARRIOR = 1125
    ELITE_KESHIK = 1230
    ELITE_KIPCHAK = 1233
    ELITE_KONNIK = 1227
    ELITE_KONNIK_DISMOUNTED = 1253
    ELITE_LEITIS = 1236
    ELITE_LONGBOAT = 533
    ELITE_LONGBOWMAN = 530
    ELITE_MAGYAR_HUSZAR = 871
    ELITE_MAMELUKE = 556
    ELITE_MANGUDAI = 561
    ELITE_ORGAN_GUN = 1003
    ELITE_PLUMED_ARCHER = 765
    ELITE_RATTAN_ARCHER = 1131
    ELITE_SAMURAI = 560
    ELITE_SHOTEL_WARRIOR = 1018
    ELITE_SKIRMISHER = 6
    ELITE_STEPPE_LANCER = 1372
    ELITE_TARKAN = 757
    ELITE_TEUTONIC_KNIGHT = 554
    ELITE_THROWING_AXEMAN = 531
    ELITE_TURTLE_SHIP = 832
    ELITE_WAR_ELEPHANT = 558
    ELITE_WAR_WAGON = 829
    ELITE_WOAD_RAIDER = 534
    FARMER = 259
    FAST_FIRE_SHIP = 532
    FIRE_GALLEY = 1103
    FIRE_SHIP = 529
    FISHERMAN = 56
    FISHING_SHIP = 13
    FLAMETHROWER = 188
    FLAMING_CAMEL = 1263
    FORAGER = 120
    FURIOUS_THE_MONKEY_BOY = 860
    GALLEON = 442
    GALLEY = 539
    GBETO = 1013
    GENITOUR = 1010
    GENOESE_CROSSBOWMAN = 866
    GOAT = 1060
    GOLD_MINER = 579
    GOOSE = 1243
    HALBERDIER = 359
    HAND_CANNONEER = 5
    HEAVY_CAMEL_RIDER = 330
    HEAVY_CAVALRY_ARCHER = 474
    HEAVY_CROSSBOWMAN = 493
    HEAVY_DEMOLITION_SHIP = 528
    HEAVY_PIKEMAN = 892
    HEAVY_SCORPION = 542
    HEAVY_SWORDSMAN = 76
    HORSE_A = 814
    HORSE_B = 1356
    HORSE_C = 1602
    HORSE_D = 1604
    HORSE_E = 1606
    HUNTER = 122
    HUSKARL = 41
    HUSSAR = 441
    IMPERIAL_CAMEL_RIDER = 207
    IMPERIAL_SKIRMISHER = 1155
    INVISIBLE_OBJECT = 1291
    IROQUOIS_WARRIOR = 1374
    JAGUAR_WARRIOR = 725
    JANISSARY = 46
    JUNK = 15
    KAMAYUK = 879
    KARAMBIT_WARRIOR = 1123
    KESHIK = 1228
    KHAN = 1275
    KING = 434
    KIPCHAK = 1231
    KNIGHT = 38
    KONNIK = 1225
    KONNIK_DISMOUNTED = 1252
    LEGIONARY = 1
    LEITIS = 1234
    LIGHT_CAVALRY = 546
    LLAMA = 305
    LONG_SWORDSMAN = 77
    LONGBOAT = 250
    LONGBOWMAN = 8
    LUMBERJACK = 123
    MAGYAR_HUSZAR = 869
    MAMELUKE = 282
    MAN_AT_ARMS = 75
    MANGONEL = 280
    MANGUDAI = 11
    MERCHANT = 1572
    MILITIA = 74
    MISSIONARY = 775
    MONK = 125
    MONK_WITH_RELIC = 286
    NINJA = 1145
    NORSE_WARRIOR = 361
    ONAGER = 550
    ORGAN_GUN = 1001
    OX_CART = 1271
    OX_WAGON = 1273
    PALADIN = 569
    PENGUIN = 639
    PETARD = 440
    PHOTONMAN = 1577
    PIG = 1245
    PIKEMAN = 358
    PLUMED_ARCHER = 763
    PRIEST = 1023
    PRIEST_WITH_RELIC = 1400
    QUEEN = 1292
    RATTAN_ARCHER = 1129
    RELIC_CART = 1304
    REPAIRER = 156
    ROYAL_JANISSARY = 52
    SAMURAI = 291
    SCORPION = 279
    SCOUT_CAVALRY = 448
    SHARKATZOR = 1222
    SHEEP = 594
    SHEPHERD = 592
    SHOTEL_WARRIOR = 1016
    SIEGE_ONAGER = 588
    SIEGE_RAM = 548
    SIEGE_TOWER = 1105
    SKIRMISHER = 7
    SLINGER = 185
    SPEARMAN = 93
    STEPPE_LANCER = 1370
    STONE_MINER = 124
    TARKAN = 755
    TEUTONIC_KNIGHT = 25
    THROWING_AXEMAN = 281
    TORCH_A_CONVERTABLE = 854
    TORCH_B_CONVERTABLE = 1377
    TRADE_CART_EMPTY = 128
    TRADE_CART_FULL = 204
    TRADE_COG = 17
    TRANSPORT_SHIP = 545
    TREBUCHET = 42
    TREBUCHET_PACKED = 331
    TURKEY = 833
    TURTLE_SHIP = 831
    TWO_HANDED_SWORDSMAN = 473
    VILLAGER_FEMALE = 293
    VILLAGER_MALE = 83
    WAR_ELEPHANT = 239
    WAR_GALLEY = 21
    WAR_WAGON = 827
    WATER_BUFFALO = 1142
    WOAD_RAIDER = 232
    XOLOTL_WARRIOR = 1570


class GaiaUnit(IntEnum):
    # GAIA Units
    BEAR = 486
    BUTTERFLY1 = 1608
    BUTTERFLY2 = 1609
    BUTTERFLY3 = 1610
    CROCODILE = 1031
    DEER = 65
    DIRE_WOLF = 89
    ELEPHANT = 1301
    FALCON = 1056
    HAWK = 96
    IBEX = 1239
    IRON_BOAR = 810
    JAGUAR = 812
    JAVELINA = 822
    KOMODO_DRAGON = 1135
    LION = 1029
    MACAW = 816
    OSTRICH = 1026
    RABID_WOLF = 202
    RHINOCEROS = 1139
    SEAGULLS = 303
    SNOW_LEOPARD = 1241
    STORK = 1028
    STORMY_DOG = 862
    TIGER = 1137
    VULTURE = 1305
    WILD_BACTRIAN_CAMEL = 1247
    WILD_BOAR = 48
    WILD_CAMEL = 884
    WILD_HORSE = 835
    WOLF = 126
    ZEBRA = 1019

    # Player Units
    MOVEABLE_MAP_REVEALER = 0
    ALFRED_THE_ALPACA = 1300
    AMAZON_ARCHER = 850
    AMAZON_WARRIOR = 825
    ARAMBAI = 1126
    ARBALESTER = 492
    ARCHER = 4
    BACTRIAN_CAMEL = 1237
    BALLISTA_ELEPHANT = 1120
    BANDIT = 299
    BATTERING_RAM = 1258
    BATTLE_ELEPHANT = 1132
    BERSERK = 692
    BOMBARD_CANNON = 36
    BOYAR = 876
    BUILDER = 118
    CAMEL = 897
    CAMEL_ARCHER = 1007
    CAMEL_RIDER = 329
    CANNON_GALLEON = 420
    CANOE = 778
    CAPPED_RAM = 422
    CARAVEL = 1004
    CART = 1338
    CATAPHRACT = 40
    CAVALIER = 283
    CAVALRY_ARCHER = 39
    CENTURION = 275
    CHAMPION = 567
    CHU_KO_NU = 73
    COBRA_CAR = 748
    CONDOTTIERO = 882
    CONQUISTADOR = 771
    COW_A = 705
    COW_B = 1596
    COW_C = 1598
    COW_D = 1600
    CROSSBOWMAN = 24
    DEMOLITION_RAFT = 1104
    DEMOLITION_SHIP = 527
    DONKEY = 846
    DRAGON_SHIP = 1302
    EAGLE_SCOUT = 751
    EAGLE_WARRIOR = 753
    EASTERN_SWORDSMAN = 894
    ELEPHANT_ARCHER = 873
    ELITE_ARAMBAI = 1128
    ELITE_BALLISTA_ELEPHANT = 1122
    ELITE_BATTLE_ELEPHANT = 1134
    ELITE_BERSERK = 694
    ELITE_BOYAR = 878
    ELITE_CAMEL_ARCHER = 1009
    ELITE_CANNON_GALLEON = 691
    ELITE_CARAVEL = 1006
    ELITE_CATAPHRACT = 553
    ELITE_CHU_KO_NU = 559
    ELITE_CONQUISTADOR = 773
    ELITE_EAGLE_WARRIOR = 752
    ELITE_ELEPHANT_ARCHER = 875
    ELITE_GBETO = 1015
    ELITE_GENITOUR = 1012
    ELITE_GENOESE_CROSSBOWMAN = 868
    ELITE_HUSKARL = 555
    ELITE_JAGUAR_WARRIOR = 726
    ELITE_JANISSARY = 557
    ELITE_KAMAYUK = 881
    ELITE_KARAMBIT_WARRIOR = 1125
    ELITE_KESHIK = 1230
    ELITE_KIPCHAK = 1233
    ELITE_KONNIK = 1227
    ELITE_KONNIK_DISMOUNTED = 1253
    ELITE_LEITIS = 1236
    ELITE_LONGBOAT = 533
    ELITE_LONGBOWMAN = 530
    ELITE_MAGYAR_HUSZAR = 871
    ELITE_MAMELUKE = 556
    ELITE_MANGUDAI = 561
    ELITE_ORGAN_GUN = 1003
    ELITE_PLUMED_ARCHER = 765
    ELITE_RATTAN_ARCHER = 1131
    ELITE_SAMURAI = 560
    ELITE_SHOTEL_WARRIOR = 1018
    ELITE_SKIRMISHER = 6
    ELITE_STEPPE_LANCER = 1372
    ELITE_TARKAN = 757
    ELITE_TEUTONIC_KNIGHT = 554
    ELITE_THROWING_AXEMAN = 531
    ELITE_TURTLE_SHIP = 832
    ELITE_WAR_ELEPHANT = 558
    ELITE_WAR_WAGON = 829
    ELITE_WOAD_RAIDER = 534
    FARMER = 259
    FAST_FIRE_SHIP = 532
    FIRE_GALLEY = 1103
    FIRE_SHIP = 529
    FISHERMAN = 56
    FISHING_SHIP = 13
    FLAMETHROWER = 188
    FLAMING_CAMEL = 1263
    FORAGER = 120
    FURIOUS_THE_MONKEY_BOY = 860
    GALLEON = 442
    GALLEY = 539
    GBETO = 1013
    GENITOUR = 1010
    GENOESE_CROSSBOWMAN = 866
    GOAT = 1060
    GOLD_MINER = 579
    GOOSE = 1243
    HALBERDIER = 359
    HAND_CANNONEER = 5
    HEAVY_CAMEL_RIDER = 330
    HEAVY_CAVALRY_ARCHER = 474
    HEAVY_CROSSBOWMAN = 493
    HEAVY_DEMOLITION_SHIP = 528
    HEAVY_PIKEMAN = 892
    HEAVY_SCORPION = 542
    HEAVY_SWORDSMAN = 76
    HORSE_A = 814
    HORSE_B = 1356
    HORSE_C = 1602
    HORSE_D = 1604
    HORSE_E = 1606
    HUNTER = 122
    HUSKARL = 41
    HUSSAR = 441
    IMPERIAL_CAMEL_RIDER = 207
    IMPERIAL_SKIRMISHER = 1155
    INVISIBLE_OBJECT = 1291
    IROQUOIS_WARRIOR = 1374
    JAGUAR_WARRIOR = 725
    JANISSARY = 46
    JUNK = 15
    KAMAYUK = 879
    KARAMBIT_WARRIOR = 1123
    KESHIK = 1228
    KHAN = 1275
    KING = 434
    KIPCHAK = 1231
    KNIGHT = 38
    KONNIK = 1225
    KONNIK_DISMOUNTED = 1252
    LEGIONARY = 1
    LEITIS = 1234
    LIGHT_CAVALRY = 546
    LLAMA = 305
    LONG_SWORDSMAN = 77
    LONGBOAT = 250
    LONGBOWMAN = 8
    LUMBERJACK = 123
    MAGYAR_HUSZAR = 869
    MAMELUKE = 282
    MAN_AT_ARMS = 75
    MANGONEL = 280
    MANGUDAI = 11
    MERCHANT = 1572
    MILITIA = 74
    MISSIONARY = 775
    MONK = 125
    MONK_WITH_RELIC = 286
    NINJA = 1145
    NORSE_WARRIOR = 361
    ONAGER = 550
    ORGAN_GUN = 1001
    OX_CART = 1271
    OX_WAGON = 1273
    PALADIN = 569
    PENGUIN = 639
    PETARD = 440
    PHOTONMAN = 1577
    PIG = 1245
    PIKEMAN = 358
    PLUMED_ARCHER = 763
    PRIEST = 1023
    PRIEST_WITH_RELIC = 1400
    QUEEN = 1292
    RATTAN_ARCHER = 1129
    RELIC_CART = 1304
    REPAIRER = 156
    ROYAL_JANISSARY = 52
    SAMURAI = 291
    SCORPION = 279
    SCOUT_CAVALRY = 448
    SHARKATZOR = 1222
    SHEEP = 594
    SHEPHERD = 592
    SHOTEL_WARRIOR = 1016
    SIEGE_ONAGER = 588
    SIEGE_RAM = 548
    SIEGE_TOWER = 1105
    SKIRMISHER = 7
    SLINGER = 185
    SPEARMAN = 93
    STEPPE_LANCER = 1370
    STONE_MINER = 124
    TARKAN = 755
    TEUTONIC_KNIGHT = 25
    THROWING_AXEMAN = 281
    TORCH_A_CONVERTABLE = 854
    TORCH_B_CONVERTABLE = 1377
    TRADE_CART_EMPTY = 128
    TRADE_CART_FULL = 204
    TRADE_COG = 17
    TRANSPORT_SHIP = 545
    TREBUCHET = 42
    TREBUCHET_PACKED = 331
    TURKEY = 833
    TURTLE_SHIP = 831
    TWO_HANDED_SWORDSMAN = 473
    VILLAGER_FEMALE = 293
    VILLAGER_MALE = 83
    WAR_ELEPHANT = 239
    WAR_GALLEY = 21
    WAR_WAGON = 827
    WATER_BUFFALO = 1142
    WOAD_RAIDER = 232
    XOLOTL_WARRIOR = 1570


unit_names = bidict({
    0: "moveable_map_revealer",
    1: "legionary",
    4: "archer",
    5: "hand_cannoneer",
    6: "elite_skirmisher",
    7: "skirmisher",
    8: "longbowman",
    11: "mangudai",
    13: "fishing_ship",
    15: "junk",
    17: "trade_cog",
    21: "war_galley",
    24: "crossbowman",
    25: "teutonic_knight",
    36: "bombard_cannon",
    38: "knight",
    39: "cavalry_archer",
    40: "cataphract",
    41: "huskarl",
    42: "trebuchet",
    46: "janissary",
    52: "royal_janissary",
    56: "fisherman",
    73: "chu_ko_nu",
    74: "militia",
    75: "man_at_arms",
    76: "heavy_swordsman",
    77: "long_swordsman",
    83: "villager_male",
    93: "spearman",
    118: "builder",
    120: "forager",
    122: "hunter",
    123: "lumberjack",
    124: "stone_miner",
    125: "monk",
    128: "trade_cart_empty",
    156: "repairer",
    185: "slinger",
    188: "flamethrower",
    204: "trade_cart_full",
    207: "imperial_camel_rider",
    232: "woad_raider",
    239: "war_elephant",
    250: "longboat",
    259: "farmer",
    275: "centurion",
    279: "scorpion",
    280: "mangonel",
    281: "throwing_axeman",
    282: "mameluke",
    283: "cavalier",
    286: "monk_with_relic",
    291: "samurai",
    293: "villager_female",
    299: "bandit",
    305: "llama",
    329: "camel_rider",
    330: "heavy_camel_rider",
    331: "trebuchet_packed",
    358: "pikeman",
    359: "halberdier",
    361: "norse_warrior",
    420: "cannon_galleon",
    422: "capped_ram",
    434: "king",
    440: "petard",
    441: "hussar",
    442: "galleon",
    448: "scout_cavalry",
    473: "two_handed_swordsman",
    474: "heavy_cavalry_archer",
    492: "arbalester",
    493: "heavy_crossbowman",
    527: "demolition_ship",
    528: "heavy_demolition_ship",
    529: "fire_ship",
    530: "elite_longbowman",
    531: "elite_throwing_axeman",
    532: "fast_fire_ship",
    533: "elite_longboat",
    534: "elite_woad_raider",
    539: "galley",
    542: "heavy_scorpion",
    545: "transport_ship",
    546: "light_cavalry",
    548: "siege_ram",
    550: "onager",
    553: "elite_cataphract",
    554: "elite_teutonic_knight",
    555: "elite_huskarl",
    556: "elite_mameluke",
    557: "elite_janissary",
    558: "elite_war_elephant",
    559: "elite_chu_ko_nu",
    560: "elite_samurai",
    561: "elite_mangudai",
    567: "champion",
    569: "paladin",
    579: "gold_miner",
    588: "siege_onager",
    592: "shepherd",
    594: "sheep",
    639: "penguin",
    691: "elite_cannon_galleon",
    692: "berserk",
    694: "elite_berserk",
    705: "cow_a",
    725: "jaguar_warrior",
    726: "elite_jaguar_warrior",
    748: "cobra_car",
    751: "eagle_scout",
    752: "elite_eagle_warrior",
    753: "eagle_warrior",
    755: "tarkan",
    757: "elite_tarkan",
    763: "plumed_archer",
    765: "elite_plumed_archer",
    771: "conquistador",
    773: "elite_conquistador",
    775: "missionary",
    778: "canoe",
    814: "horse_a",
    825: "amazon_warrior",
    827: "war_wagon",
    829: "elite_war_wagon",
    831: "turtle_ship",
    832: "elite_turtle_ship",
    833: "turkey",
    846: "donkey",
    850: "amazon_archer",
    854: "torch_a_convertable",
    860: "furious_the_monkey_boy",
    866: "genoese_crossbowman",
    868: "elite_genoese_crossbowman",
    869: "magyar_huszar",
    871: "elite_magyar_huszar",
    873: "elephant_archer",
    875: "elite_elephant_archer",
    876: "boyar",
    878: "elite_boyar",
    879: "kamayuk",
    881: "elite_kamayuk",
    882: "condottiero",
    892: "heavy_pikeman",
    894: "eastern_swordsman",
    897: "camel",
    1001: "organ_gun",
    1003: "elite_organ_gun",
    1004: "caravel",
    1006: "elite_caravel",
    1007: "camel_archer",
    1009: "elite_camel_archer",
    1010: "genitour",
    1012: "elite_genitour",
    1013: "gbeto",
    1015: "elite_gbeto",
    1016: "shotel_warrior",
    1018: "elite_shotel_warrior",
    1023: "priest",
    1060: "goat",
    1103: "fire_galley",
    1104: "demolition_raft",
    1105: "siege_tower",
    1120: "ballista_elephant",
    1122: "elite_ballista_elephant",
    1123: "karambit_warrior",
    1125: "elite_karambit_warrior",
    1126: "arambai",
    1128: "elite_arambai",
    1129: "rattan_archer",
    1131: "elite_rattan_archer",
    1132: "battle_elephant",
    1134: "elite_battle_elephant",
    1142: "water_buffalo",
    1145: "ninja",
    1155: "imperial_skirmisher",
    1222: "sharkatzor",
    1225: "konnik",
    1227: "elite_konnik",
    1228: "keshik",
    1230: "elite_keshik",
    1231: "kipchak",
    1233: "elite_kipchak",
    1234: "leitis",
    1236: "elite_leitis",
    1237: "bactrian_camel",
    1243: "goose",
    1245: "pig",
    1252: "konnik_dismounted",
    1253: "elite_konnik_dismounted",
    1258: "battering_ram",
    1263: "flaming_camel",
    1271: "ox_cart",
    1273: "ox_wagon",
    1275: "khan",
    1291: "invisible_object",
    1292: "queen",
    1300: "alfred_the_alpaca",
    1302: "dragon_ship",
    1304: "relic_cart",
    1338: "cart",
    1356: "horse_b",
    1370: "steppe_lancer",
    1372: "elite_steppe_lancer",
    1374: "iroquois_warrior",
    1377: "torch_b_convertable",
    1400: "priest_with_relic",
    1570: "xolotl_warrior",
    1572: "merchant",
    1577: "photonman",
    1596: "cow_b",
    1598: "cow_c",
    1600: "cow_d",
    1602: "horse_c",
    1604: "horse_d",
    1606: "horse_e",

    # Gaia Units
    48: "wild_boar",
    65: "deer",
    89: "dire_wolf",
    96: "hawk",
    126: "wolf",
    202: "rabid_wolf",
    303: "seagulls",
    486: "bear",
    810: "iron_boar",
    812: "jaguar",
    816: "macaw",
    822: "javelina",
    835: "wild_horse",
    862: "stormy_dog",
    884: "wild_camel",
    1019: "zebra",
    1026: "ostrich",
    1028: "stork",
    1029: "lion",
    1031: "crocodile",
    1137: "tiger",
    1056: "falcon",
    1135: "komodo_dragon",
    1139: "rhinoceros",
    1241: "snow_leopard",
    1247: "wild_bactrian_camel",
    1301: "elephant",
    1305: "vulture",
    1239: "ibex",
    1608: "butterfly1",
    1609: "butterfly2",
    1610: "butterfly3",
})



import AoE2ScenarioParser.pieces.structs.aoe2_struct
from AoE2ScenarioParser.helper.bytes_to_x import *
from AoE2ScenarioParser.helper.generator import repeat_generator as r_gen

types = [
    "s",  # Signed int
    "u",  # Unsigned int
    "f",  # FloatingPoint
    "c",  # Character string
    "str",  # Variable length string
    "data",  # Data (Can be changed by used using bytes_to_x functions)
]


def vorl(var):
    """vorl stands for "Variable or List". This function returns the value if the list is a size of 1"""
    if type(var) is list:
        if len(var) is 1:
            return var[0]
        else:
            return var
    else:
        return var


def listify(var):
    """Always return item as list"""
    if type(var) is list:
        return var
    else:
        return [var]


class Parser:
    def __init__(self):
        self._saves = dict()

    def retrieve_value(self, generator, retriever, as_length=False):
        length = 0
        result = list()
        var_type, var_len = datatype_to_type_length(retriever.datatype.var)

        if retriever.set_repeat is not None:
            retriever.datatype.repeat = parse_repeat_string(self._saves, retriever.set_repeat)
        # try:
        for i in range(0, retriever.datatype.repeat):
            length += var_len

            if var_type == "struct":
                val = retriever.datatype.var(self)
                val.set_data_from_generator(generator)
                result.append(val)
                i = val.get_length()

                length += i
                continue
            if var_type == "u" or var_type == "s":
                val = bytes_to_int(r_gen(generator, var_len), signed=(var_type == "s"))
            elif var_type == "f":
                if var_len == 4:
                    val = bytes_to_float(r_gen(generator, var_len))
                else:  # Always 4 except for trigger version
                    val = bytes_to_double(r_gen(generator, var_len))
            elif var_type == "c":
                val = bytes_to_str(r_gen(generator, var_len))
            elif var_type == "data":
                val = r_gen(generator, var_len)
            elif var_type == "str":
                string_length = bytes_to_int(r_gen(generator, var_len), endian="little", signed=True)
                try:
                    data = r_gen(generator, string_length)
                    val = bytes_to_str(data)
                    length += string_length
                except StopIteration as e:
                    print(f"\n[StopIteration] Parser.retrieve_value: \n"
                          f"\tRetriever: {retriever}\n"
                          f"\tString length: {string_length}\n")
                    raise StopIteration(e)
            else:
                break

            result.append(val)
        # except StopIteration as e:
        #     print(repr(retriever) + "\n" +
        #           "var_type: " + var_type + "\n" +
        #           "var_len: " + str(var_len))
        #     print(e)
        #     exit()

        if retriever.on_success is not None:
            if type(result) is list:
                for x in range(0, len(result)):
                    result[x] = retriever.on_success(result[x])
            else:
                result = retriever.on_success(result)

        if retriever.save_as is not None:
            self.add_to_saves(retriever.save_as, vorl(result))

        if retriever.log_value:
            print(retriever, "retrieved:", vorl(result))

        return vorl(result) if not as_length else length

    def add_to_saves(self, name, value):
        self._saves[name] = value


def parse_repeat_string(saves, repeat_string):
    while True:
        start = repeat_string.find("{")
        end = repeat_string.find("}")

        if start is -1 and end is -1:
            break

        inclusive = repeat_string[start:end + 1]
        exclusive = repeat_string[start + 1:end]

        repeat_string = repeat_string.replace(inclusive, str(saves[exclusive]))
    return eval(repeat_string)


def calculate_length(generator, retriever_list):
    parser = Parser()
    total_length = 0

    for retriever in retriever_list:
        total_length += parser.retrieve_value(generator, retriever, as_length=True)

    return total_length


def datatype_to_type_length(var):
    try:
        if issubclass(var, AoE2ScenarioParser.pieces.structs.aoe2_struct.AoE2Struct):
            return "struct", 0
        else:  # Not possible at this time
            return "", 0
    except TypeError:
        pass

    var_type = ""
    var_len = ""

    for char in var:
        if char.isnumeric():
            var_len += char
        else:
            var_type += char

    if var_type == "":
        var_type = "data"

    if var_len == "":
        var_len = 0
    else:
        var_len = int(var_len)

    assert var_type in types

    if var_type != "c" and var_type != "data":
        var_len = int(var_len / 8)

    return var_type, var_len


def retriever_to_bytes(retriever):
    var_type, var_len = datatype_to_type_length(retriever.datatype.var)

    return_bytes = b''

    is_list = type(retriever.data) == list
    if is_list:
        retriever.datatype.repeat = len(retriever.data)

    try:
        for i in range(0, retriever.datatype.repeat):
            data = retriever.data[i] if is_list else retriever.data

            if data is None:
                print("No data found in retriever: " + str(retriever))
                return None

            if var_type == "struct":
                for struct_retriever in data.retrievers:
                    return_bytes += retriever_to_bytes(struct_retriever)
            if var_type == "u" or var_type == "s":
                return_bytes += int_to_bytes(data, var_len, signed=(var_type == "s"))
            elif var_type == "f":
                if var_len == 4:
                    return_bytes += float_to_bytes(data)
                else:  # Always 4 (float) except for trigger version (8: double)
                    return_bytes += double_to_bytes(data)
            elif var_type == "c":
                return_bytes += str_to_bytes(data)
            elif var_type == "data":
                return_bytes += data
            elif var_type == "str":
                return_bytes += int_to_bytes(len(data.encode('utf-8')), var_len, endian="little", signed=True)
                return_bytes += str_to_bytes(data)
    except (AttributeError, TypeError) as e:
        print("\n" + type(e).__name__ + " occurred in: " + retriever.name +
              "\n\tData: " + repr(retriever.data) + "\n\tDatatype: " + str(retriever.datatype))
        raise e

    if retriever.log_value:
        print(retriever, "returned", return_bytes)

    return return_bytes



from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.datatype import DataType


class OptionsPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Per player number of disabled techs", DataType("u32", repeat=16), save_as="disabled_techs"),
            Retriever("Disabled technology IDs in player order", DataType("u32"), set_repeat="sum({disabled_techs})"),
            # Retriever("Per player extra number of disabled techs", DataType("u32", repeat=16)),  # Removed in DE?
            Retriever("Per player number of disabled units", DataType("u32", repeat=16), save_as="disabled_units"),
            Retriever("Disabled unit IDs in player order", DataType("u32"), set_repeat="sum({disabled_units})"),
            # Retriever("Per player extra number of disabled units", DataType("u32", repeat=16)),  # Removed in DE?
            Retriever("Per player number of disabled buildings",
                      DataType("u32", repeat=16), save_as="disabled_buildings"),
            Retriever("Disabled building IDs in player order",
                      DataType("u32"), set_repeat="sum({disabled_buildings})"),
            Retriever("Combat Mode", DataType("u32")),
            Retriever("Naval Mode", DataType("u32")),
            Retriever("All techs", DataType("u32")),
            Retriever("Per player starting age", DataType("u32", repeat=16)),  # 2: Dark 6 = Post | 1-8 players 9 GAIA
            Retriever("Unknown", DataType("36")),
        ]

        super().__init__("Options", retrievers, parser_obj, data=data)



from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.datatype import DataType


class BackgroundImagePiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("ASCII, Background filename", DataType("str16")),
            Retriever("Picture Version", DataType("u32")),
            Retriever("Bitmap width", DataType("u32")),
            Retriever("Bitmap height", DataType("s32")),
            Retriever("Picture Orientation", DataType("s16")),
            # Retriever("	BITMAPINFOHEADER", DataType("u32")),
        ]

        super().__init__("Background Image", retrievers, parser_obj, data=data)



import pyautogui
from time import sleep
from tkinter import Tk
r = Tk()


"""
When copying select the top entry. With Units and Buildings this'll be '<none>'. For technologies this isn't the case. 
So just select the top Technology.

For units/buildings: Use 'Create Object'
For Technologies: Use 'Enable/Disable Technology'
"""
copying_techs = False


def copy_and_print(last=""):
    pyautogui.hotkey('ctrl', 'c')
    obj_id = r.selection_get(selection="CLIPBOARD")
    a = check_if_duplicate(obj_id, last)
    print(obj_id)
    return a


def check_if_duplicate(new, last):
    if new == last:
        exit("Same ID twice. End reached")
    else:
        return new


for x in range(5, 0, -1):
    print(x, end=" ")
    sleep(1)
print("\n")

# Change to 'tabbing mode'
pyautogui.press('tab')
sleep(0.2)

last_obj_id = ""

if copying_techs:
    pyautogui.hotkey('shiftleft', 'tab')
    sleep(0.05)
    pyautogui.press('enter')
    sleep(0.05)
    pyautogui.press('enter')
    sleep(0.05)
    pyautogui.press('tab')
    copy_and_print()

while True:
    # Move to dropdown
    pyautogui.hotkey('shiftleft', 'tab')
    if not copying_techs:
        pyautogui.hotkey('shiftleft', 'tab')
    sleep(0.05)

    # Open dropdown
    pyautogui.press('enter')
    sleep(0.05)

    # Next unit
    pyautogui.press('down')
    sleep(0.05)

    # Select unit
    pyautogui.press('enter')
    sleep(0.05)

    # Move to ID field
    pyautogui.press('tab')
    if not copying_techs:
        pyautogui.press('tab')
    sleep(0.05)

    last_obj_id = copy_and_print(last=last_obj_id)



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class ResourcesStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Gold", DataType("u32")),
            Retriever("Wood", DataType("u32")),
            Retriever("Food", DataType("u32")),
            Retriever("Stone", DataType("u32")),
            Retriever("Ore X (unused)", DataType("u32")),
            Retriever("Trade Goods", DataType("u32")),
            Retriever("Player color", DataType("u32"))
        ]

        super().__init__("Resources", retrievers, parser_obj, data)



from __future__ import annotations

from typing import List

from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper import parser
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.trigger_obj import TriggerObject
from AoE2ScenarioParser.objects.variable_obj import VariableObject


def _evaluate_index_params(trigger_id, display_index):
    if trigger_id is None and display_index is None:
        raise ValueError("Please choose 'trigger_id' or 'display_index' as identification for the wanted trigger")
    if trigger_id is not None and display_index is not None:
        raise ValueError("Please identify a trigger using 'trigger_id' or 'display_index' but not both")


class TriggersObject(AoE2Object):
    def __init__(self,
                 triggers,
                 trigger_display_order,
                 variables
                 ):

        self.triggers: List[TriggerObject] = parser.listify(triggers)
        self.trigger_display_order: List[int] = parser.listify(trigger_display_order)
        self.variables: List[VariableObject] = parser.listify(variables)

        super().__init__()

    def add_trigger(self, name) -> TriggerObject:
        new_trigger = TriggerObject(name=name, trigger_id=len(self.triggers))
        self.triggers.append(new_trigger)
        helper.update_order_array(self.trigger_display_order, len(self.triggers))
        return new_trigger

    def add_variable(self, variable_id, name) -> VariableObject:
        if not (0 <= variable_id <= 255):
            raise ValueError("Variable ID has to fall between 0 and 255 (incl).")

        new_variable = VariableObject(variable_id=variable_id, name=name)
        self.variables.append(new_variable)
        return new_variable

    def get_summary_as_string(self) -> str:
        return_string = "\nTrigger Summary:\n"

        triggers = parser.listify(self.triggers)
        display_order = parser.listify(self.trigger_display_order)

        if len(display_order) == 0:
            return_string += "\t<< No Triggers >>"

        longest_trigger_name = -1
        for trigger_id in display_order:
            trigger_name = helper.del_str_trail(triggers[trigger_id].name)
            longest_trigger_name = max(longest_trigger_name, len(trigger_name))

        longest_trigger_name += 3
        for display, trigger_id in enumerate(display_order):
            trigger = triggers[trigger_id]
            trigger_name = helper.del_str_trail(trigger.name)

            buffer = longest_trigger_name - len(trigger_name)
            return_string += "\t" + trigger_name + (" " * buffer)
            return_string += " [Index: " + str(trigger_id) + ", Display: " + str(display) + "]"

            return_string += "\t(conditions: " + str(len(parser.listify(trigger.conditions))) + ", "
            return_string += " effects: " + str(len(parser.listify(trigger.effects))) + ")\n"

        variables = parser.listify(self.variables)

        return_string += "\nVariables Summary:\n"
        if len(variables) == 0:
            return_string += "\t<< No Variables >>"

        longest_variable_name = -1
        for variable in variables:
            longest_variable_name = max(longest_variable_name, len(variable.name))

        longest_variable_name += 3
        for index, variable in enumerate(variables):
            var_name = variable.name
            buffer = " " * (longest_variable_name - len(var_name))
            return_string += f"\t{var_name}{buffer}[Index: {variable.variable_id}]\n"

        return return_string

    def get_content_as_string(self) -> str:
        return_string = "Triggers:\n"

        triggers = parser.listify(self.triggers)
        display_order = parser.listify(self.trigger_display_order)

        if len(triggers) == 0:
            return_string += "\t<<No triggers>>"

        for trigger_id in display_order:
            return_string += self.get_trigger_as_string(trigger_id) + "\n"

        return_string += "Variables:\n"

        if len(triggers) == 0:
            return_string += "\t<<No Variables>>"

        for variable in parser.listify(self.variables):
            return_string += f"\t'{variable.name}' [Index: {variable.variable_id}]\n"

        return return_string

    def get_trigger_as_string(self, trigger_id=None, display_index=None) -> str:
        _evaluate_index_params(trigger_id, display_index)

        if trigger_id is None:
            trigger_id = self._get_trigger_id_by_display_index(display_index)

        trigger: TriggerObject = parser.listify(self.triggers)[trigger_id]
        display = parser.listify(self.trigger_display_order).index(trigger_id)

        return_string = "\t'" + helper.del_str_trail(trigger.name) + "'"
        return_string += " [Index: " + str(trigger_id) + ", Display: " + str(display) + "]" + ":\n"

        return_string += trigger.get_content_as_string()

        return return_string

    def get_trigger(self, trigger_id=None, display_index=None) -> TriggerObject:
        _evaluate_index_params(trigger_id, display_index)

        if trigger_id is None:
            trigger_id = self._get_trigger_id_by_display_index(display_index)

        return parser.listify(self.triggers)[trigger_id]

    def get_variable(self, variable_id: int = None, variable_name: str = None) -> VariableObject:
        if variable_id is None and variable_name is None:
            raise ValueError("Select a variable using the variable_id or variable_name parameters")
        if variable_id is not None and variable_name is not None:
            raise ValueError("Select a variable using either the variable_id or variable_name parameters, not both.")

        for variable in self.variables:
            if variable.variable_id == variable_id or variable.name == variable_name:
                return variable

    def remove_trigger(self, trigger_id: int = None, display_index: int = None, trigger: TriggerObject = None) -> None:
        if trigger is None:
            _evaluate_index_params(trigger_id, display_index)
        else:
            trigger_id = trigger.trigger_id

        if trigger_id is None:
            trigger_id = self._get_trigger_id_by_display_index(display_index)
        else:
            display_index = self.trigger_display_order.index(trigger_id)

        del self.triggers[trigger_id]
        del self.trigger_display_order[display_index]

        self.trigger_display_order = \
            [x - 1 if x > trigger_id else x for x in self.trigger_display_order]

    def _get_trigger_id_by_display_index(self, display_index) -> int:
        return self.trigger_display_order[display_index]

    @staticmethod
    def _parse_object(parsed_data, **kwargs) -> TriggersObject:  # Expected {}
        display_order = parser.listify(
            find_retriever(parsed_data['TriggerPiece'].retrievers, "Trigger display order array").data)
        trigger_data = parser.listify(
            find_retriever(parsed_data['TriggerPiece'].retrievers, "Trigger data").data)
        var_data = parser.listify(
            find_retriever(parsed_data['TriggerPiece'].retrievers, "Variables").data)

        triggers = []
        for index, trigger in enumerate(trigger_data):
            triggers.append(TriggerObject._parse_object(parsed_data, trigger=trigger, trigger_id=index))

        variables = []
        for var in var_data:
            variables.append(VariableObject._parse_object(parsed_data, variable=var))

        return TriggersObject(
            triggers=triggers,
            trigger_display_order=display_order,
            variables=variables
        )

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs) -> None:  # Expected {}
        number_of_triggers_retriever = find_retriever(parsed_data['TriggerPiece'].retrievers, "Number of triggers")
        trigger_data_retriever = find_retriever(parsed_data['TriggerPiece'].retrievers, "Trigger data")
        display_order_retriever = find_retriever(parsed_data['TriggerPiece'].retrievers, "Trigger display order array")
        display_order_retriever.data = parser.listify(display_order_retriever.data)
        file_header_trigger_count_retriever = find_retriever(parsed_header['FileHeaderPiece'].retrievers,
                                                             "Trigger count")

        number_of_variable_retriever = find_retriever(
            parsed_data['TriggerPiece'].retrievers, "Number of variables")
        variable_data_retriever = find_retriever(parsed_data['TriggerPiece'].retrievers, "Variables")

        trigger_data_retriever.data = []
        for trigger in objects["TriggersObject"].triggers:
            TriggerObject._reconstruct_object(parsed_header, parsed_data, objects, trigger=trigger)

        variable_data_retriever.data = []
        for variable in objects["TriggersObject"].variables:
            VariableObject._reconstruct_object(parsed_header, parsed_data, objects, variable=variable,
                                               variables=variable_data_retriever.data)

        assert len(trigger_data_retriever.data) == len(objects["TriggersObject"].triggers)
        trigger_count = len(trigger_data_retriever.data)
        number_of_triggers_retriever.data = trigger_count
        file_header_trigger_count_retriever.data = trigger_count
        number_of_variable_retriever.data = len(variable_data_retriever.data)
        # Currently not necessary due to the parser changing the 'repeated' value equal to the len(list)
        # trigger_data_retriever.datatype.repeat = trigger_count
        # display_order_retriever.datatype.repeat = trigger_count
        helper.update_order_array(display_order_retriever.data, trigger_count)



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class FileHeaderObject(AoE2Object):
    def __init__(self,
                 version,
                 timestamp,
                 instructions,
                 player_count,
                 creator_name,
                 ):

        self.version = version
        self.timestamp = timestamp
        self.instructions = instructions
        self.player_count = player_count
        self.creator_name = creator_name

        super().__init__()

    @staticmethod
    def _parse_object(parsed_data, **kwargs):
        pass

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs):
        pass


# For when turning it back into bytes:
# import time
# timestamp = int(time.time()) <-- for seconds since epoch



import struct
from enum import Enum, IntEnum


def bytes_to_str(byte_elements, codec="utf-8"):
    return byte_elements.decode(codec)


def str_to_bytes(string, codec="utf-8"):
    return string.encode(codec)


def bytes_to_int(byte_elements, endian="little", signed=False):
    return int.from_bytes(byte_elements, endian, signed=signed)


def int_to_bytes(integer: IntEnum, length, endian="little", signed=True):
    if isinstance(integer, IntEnum):
        integer = integer.value
    return integer.to_bytes(length, byteorder=endian, signed=signed)


def bytes_to_float(byte_elements):
    return struct.unpack('f', byte_elements)[0]


def float_to_bytes(f):
    return struct.pack('f', f)


def bytes_to_double(byte_elements):
    return struct.unpack('d', byte_elements)[0]


def double_to_bytes(d):
    return struct.pack('d', d)



from AoE2ScenarioParser.objects.aoe2_object import AoE2Object


class OptionsObject(AoE2Object):
    def __init__(self,
                 disables,
                 all_techs
                 ):

        self.disables = disables
        self.all_techs = all_techs

        super().__init__()

    @staticmethod
    def _parse_object(parsed_data, **kwargs):
        pass

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs):
        pass



from __future__ import annotations

from typing import List

from AoE2ScenarioParser.datasets.players import Player
from AoE2ScenarioParser.helper import parser
from AoE2ScenarioParser.helper.helper import Tile
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.objects.unit_obj import UnitObject
from AoE2ScenarioParser.pieces.structs.player_units import PlayerUnitsStruct


class UnitsObject(AoE2Object):
    def __init__(self,
                 units: List[List[UnitObject]]
                 ):

        self.units = units

        super().__init__()

    def add_unit(self, player: Player, unit_id: int, x: float, y: float, z: float = 0, rotation: float = 0,
                 garrisoned_in_id: int = -1, animation_frame: int = 0, status: int = 2,
                 reference_id: int = None, ) -> UnitObject:
        """
        Adds a unit to the scenario.

        Args:
            player: The player the unit belongs to.
            unit_id: Defines what unit you're placing. The IDs used in the unit/buildings dataset.
            x: The x location in the scenario.
            y: The y location in the scenario.
            z: The z (height) location in the scenario.
            rotation: The rotation of the unit.
            garrisoned_in_id: The reference_id of another unit this unit is garrisoned in.
            animation_frame: The animation frame of the unit.
            status: Unknown - Always 2. 0-6 no difference (?) | 7-255 makes it disappear. (Except from the mini-map)
            reference_id: The reference ID of this unit. Normally added automatically. Used for garrisoning or reference
                in triggers
        Returns:
            The UnitObject created
        """
        if reference_id is None:
            reference_id = self.get_new_reference_id()

        unit = UnitObject(
            player=player,
            x=x,
            y=y,
            z=z,
            reference_id=reference_id,
            unit_const=unit_id,
            status=status,
            rotation=rotation,
            animation_frame=animation_frame,
            garrisoned_in_id=garrisoned_in_id,
        )

        self.units[player.value].append(unit)
        return unit

    def get_player_units(self, player: Player) -> List[UnitObject]:
        """
        Returns a list of UnitObjects for the given player.

        Raises:
            ValueError: If player is not between 0 (GAIA) and 8 (EIGHT)
        """
        if not 0 <= player.value <= 8:
            raise ValueError("Player must have a value between 0 and 8")
        return self.units[player.value]

    def get_all_units(self) -> List[UnitObject]:
        units = []
        for player_units in self.units:
            units += player_units
        return units

    def get_units_in_area(self, x1: float = None, y1: float = None, x2: float = None, y2: float = None,
                          tile1: Tile = None, tile2: Tile = None, unit_list: List[UnitObject] = None,
                          players: List[Player] = None, ignore_players: List[Player] = None):
        """
        Returns all units in the square with left corner (x1, y1) and right corner (x2, y2). Both corners inclusive.

        Args:
            x1: The X location of the left corner
            y1: The Y location of the left corner
            x2: The X location of the right corner
            y2: The Y location of the right corner
            tile1: The x,y location of the 1st corner as Tile Object
            tile2: The x,y location of the 2nd corner as Tile Object
            unit_list: (Optional) A list of units (Defaults to all units in the map, including GAIA (Trees etc.)
            players: (Optional) A list of Players which units need to be selected from the selected area
            ignore_players: (Optional) A list of Players which units need to be ignored from the selected area

        Raises:
            ValueError: if not all 4 (x1, y1, x2 and y2) are used simultaneously.
                Or if both (tile1 and tile2) are not used simultaneously.
                Or if any of the 4 (x1, y1, x2, y2) is used together with any of (tile1, tile2). Use one or the other.
                Or if players and ignore_players are used simultaneously.

        :Authors:
            KSneijders (https://github.com/KSneijders/)
            T-West (https://github.com/twestura/)
        """
        if (x1 is not None or y1 is not None or x2 is not None or y2 is not None) and any([tile1, tile2]):
            raise ValueError("Cannot use both x1,y1,x2,y2 notation and tile1,tile2 notation at the same time")
        if (x1 is not None or y1 is not None or x2 is not None or y2 is not None) and \
                (x1 is None or y1 is None or x2 is None or y2 is None):
            raise ValueError("Cannot use some but not all from x1,y1,x2,y2.")
        if (not all([tile1, tile2])) and any([tile1, tile2]):
            raise ValueError("Cannot use one from tile1, tile2. Use both.")
        if players is not None and ignore_players is not None:
            raise ValueError("Cannot use both whitelist and blacklist at the same time")

        if tile1:
            x1 = tile1.x1
            y1 = tile1.y1
            x2 = tile2.x2
            y2 = tile2.y2

        if players is not None:
            players = players
        elif ignore_players is not None:
            players = [p for p in Player if p not in ignore_players]
        else:
            players = [p for p in Player]

        if unit_list is None:
            unit_list = self.get_all_units()

        return [unit for unit in unit_list
                if x1 <= unit.x <= x2 and y1 <= unit.y <= y2 and unit.player in players]

    def remove_eye_candy(self) -> None:
        eye_candy_ids = [1351, 1352, 1353, 1354, 1355, 1358, 1359, 1360, 1361, 1362, 1363, 1364, 1365, 1366]
        self.units[0] = [gaia_unit for gaia_unit in self.units[0] if gaia_unit.unit_id not in eye_candy_ids]

    def change_ownership(self, unit: UnitObject, to_player: Player) -> None:
        """
        Changes a unit's ownership to the given player.

        Args:
            unit: The unit object which ownership will be changed
            to_player: The player that'll get ownership over the unit (using Player enum)
        """
        for i, player_unit in enumerate(self.units[unit.player.value]):
            if player_unit == unit:
                del self.units[unit.player.value][i]
                self.units[to_player.value].append(unit)
                unit._player = Player(to_player)
                return

    def get_new_reference_id(self) -> int:
        highest_id = 0  # If no units, default to 0
        for player in range(0, 9):
            for unit in self.units[player]:
                if highest_id < unit.reference_id:
                    highest_id = unit.reference_id
        return highest_id + 1

    def remove_unit(self, reference_id: int = None, unit: UnitObject = None):
        if reference_id is not None and unit is not None:
            raise ValueError("Cannot use both unit_ref_id and unit arguments. Use one or the other.")
        if reference_id is None and unit is None:
            raise ValueError("Both unit_ref_id and unit arguments were unused. Use one.")

        if reference_id is not None:
            for player in range(0, 9):
                for i, unit in enumerate(self.units[player]):
                    if unit.reference_id == reference_id:
                        del self.units[player][i]
        elif unit is not None:
            self.units[unit.player.value].remove(unit)

    @staticmethod
    def _parse_object(parsed_data, **kwargs) -> UnitsObject:
        object_piece = parsed_data['UnitsPiece']
        units_per_player = find_retriever(object_piece.retrievers, "Player Units").data

        player_units = []
        for player_id in range(0, 9):  # 0 Gaia & 1-8 Players:
            player_units.append([])
            units = parser.listify(find_retriever(units_per_player[player_id].retrievers, "Units").data)

            for unit in units:
                player_units[player_id].append(
                    UnitObject._parse_object(parsed_data, unit=unit, player=Player(player_id))
                )

        return UnitsObject(
            units=player_units
        )

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs) -> None:  # Expected {}
        player_units_retriever = find_retriever(parsed_data['UnitsPiece'].retrievers, "Player Units")

        # Todo: Move this to DataHeader
        new_unit_id_retriever = find_retriever(parsed_data['DataHeaderPiece'].retrievers, "Next unit ID to place")
        new_unit_id_retriever.data = objects['UnitsObject'].get_new_reference_id()

        player_units_retriever.data = []
        for player_units in objects['UnitsObject'].units:

            units_list = []
            for unit in player_units:
                UnitObject._reconstruct_object(parsed_header, parsed_data, objects, unit=unit, units=units_list)

            player_units_retriever.data.append(
                PlayerUnitsStruct(data=[len(units_list), units_list])
            )



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class PlayerDataOneStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Active", DataType("u32")),
            Retriever("Human", DataType("u32")),
            Retriever("Civilization", DataType("u32")),
            Retriever("CTY mode", DataType("u32"))
        ]

        super().__init__("Player Data #1", retrievers, parser_obj, data)



from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.datatype import DataType


class MessagesPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Instructions", DataType("u32")),
            Retriever("Hints", DataType("u32")),
            Retriever("Victory", DataType("u32")),
            Retriever("Loss", DataType("u32")),
            Retriever("History", DataType("u32")),
            Retriever("Scouts", DataType("u32")),
            Retriever("ASCII Instructions", DataType("str16")),
            Retriever("ASCII Hints", DataType("str16")),
            Retriever("ASCII Victory", DataType("str16")),
            Retriever("ASCII Loss", DataType("str16")),
            Retriever("ASCII History", DataType("str16")),
            Retriever("ASCII Scouts", DataType("str16")),
        ]

        super().__init__("Messages", retrievers, parser_obj, data=data)



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct


class ConditionStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("condition_type", DataType("s32")),
            Retriever("static_value_21", DataType("s32")),  # Was always: 0x10
            Retriever("amount_or_quantity", DataType("s32")),  # Also technology state, also Difficulty
            Retriever("resource_type_or_tribute_list", DataType("s32")),
            Retriever("unit_object", DataType("s32")),
            Retriever("next_object", DataType("s32")),  # Old: Unit ID
            Retriever("object_list", DataType("s32")),
            Retriever("player", DataType("s32")),
            Retriever("technology", DataType("s32")),
            Retriever("timer", DataType("s32")),
            Retriever("unknown", DataType("s32")),
            Retriever("area_1_x", DataType("s32")),
            Retriever("area_1_y", DataType("s32")),
            Retriever("area_2_x", DataType("s32")),
            Retriever("area_2_y", DataType("s32")),
            Retriever("object_group", DataType("s32")),
            Retriever("object_type", DataType("s32")),  # -1: None, 1: Other, 2: Building, 3: Civilian, 4: Military
            Retriever("ai_signal", DataType("s32")),
            Retriever("inverted", DataType("s32")),
            Retriever("unknown_2", DataType("s32")),
            Retriever("variable", DataType("s32")),  # Number == VariableX
            Retriever("comparison", DataType("s32")),  # 0: ==, 1: <, 2: >, 3: <=, 4 >=
            Retriever("target_player", DataType("s32")),
        ]

        super().__init__("Condition", retrievers, parser_obj, data)



from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever


class GlobalVictoryPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Separator", DataType("u32")),
            Retriever("Conquest required", DataType("u32")),
            Retriever("Ruins", DataType("u32")),
            Retriever("Artifacts", DataType("u32")),
            Retriever("Discovery", DataType("u32")),
            Retriever("Explored % of map required", DataType("u32")),
            Retriever("Gold", DataType("u32")),
            Retriever("All custom conditions required?", DataType("u32")),
            Retriever("Mode", DataType("u32")),
            Retriever("Required score for score victory", DataType("u32")),
            Retriever("Time for timed game, in 10ths of a year", DataType("u32")),
        ]

        super().__init__("Global Victory", retrievers, parser_obj, data=data)



from enum import IntEnum

from bidict import bidict


class Tech(IntEnum):
    ANARCHY = 16
    ANDEAN_SLING = 516
    ARBALESTER = 237
    ARCHITECTURE = 51
    ARQUEBUS = 573
    ARROWSLITS = 608
    ARSON = 602
    ARTILLERY = 10
    ATHEISM = 21
    ATLATL = 460
    ATONEMENT = 319
    AZTECS = 543
    BAGAINS = 686
    BALLISTICS = 93
    BANKING = 17
    BEARDED_AXE = 83
    BERBERS = 583
    BERSERKERGANG = 49
    BLAST_FURNACE = 75
    BLOCK_PRINTING = 230
    BLOODLINES = 435
    BODKIN_ARROW = 200
    BOMBARD_CANNON = 188
    BOMBARD_TOWER = 64
    BOW_SAW = 203
    BRACER = 201
    BRITONS = 529
    BULGARIANS = 673
    BURMESE = 652
    BYZANTINES = 535
    CANNON_GALLEON = 37
    CAPPED_RAM = 96
    CARAVAN = 48
    CAREENING = 374
    CARRACK = 572
    CARTOGRAPHY = 19
    CASTLE_AGE = 102
    CAVALIER = 209
    CELTS = 541
    CHAIN_BARDING_ARMOR = 82
    CHAIN_MAIL_ARMOR = 76
    CHAMPION = 264
    CHATRAS = 628
    CHEMISTRY = 47
    CHINESE = 534
    CHIVALRY = 493
    COINAGE = 23
    CONSCRIPTION = 315
    CORVINIAN_ARMY = 514
    COURIERS = 517
    CRENELLATIONS = 11
    CROP_ROTATION = 12
    CROSSBOWMAN = 100
    CHINESE_50_PERCENT_HP_DEMOS = 396
    CUMAN_MERCENARIES = 690
    CUMANS = 675
    DARK_AGE = 104
    DOUBLE_CROSSBOW = 623
    FRANKS_FREE_FARMING_1 = 287
    DOUBLE_BIT_AXE = 202
    DRILL = 6
    DRUZHINA = 513
    DRY_DOCK = 375
    EAGLE_WARRIOR = 384
    EL_DORADO = 4
    ELITE_ARAMBAI = 619
    ELITE_BALLISTA_ELEPHANT = 615
    ELITE_BATTLE_ELEPHANT = 631
    ELITE_BERSERK = 398
    ELITE_BOYAR = 504
    ELITE_CAMEL_ARCHER = 565
    ELITE_CANNON_GALLEON = 376
    ELITE_CARAVEL = 597
    ELITE_CATAPHRACT = 361
    ELITE_CHU_KO_NU = 362
    ELITE_CONQUISTADOR = 60
    ELITE_EAGLE_WARRIOR = 434
    ELITE_ELEPHANT_ARCHER = 481
    ELITE_GBETO = 567
    ELITE_GENITOUR = 599
    ELITE_GENOESE_CROSSBOWMAN = 468
    ELITE_HUSKARL = 365
    ELITE_JAGUAR_WARRIOR = 432
    ELITE_JANISSARY = 369
    ELITE_KAMAYUK = 509
    ELITE_KARAMBIT_WARRIOR = 617
    ELITE_KESHIK = 680
    ELITE_KIPCHAK = 682
    CHIEFTAINS = 463
    ELITE_KONNIK = 678
    ELITE_LEITIS = 684
    ELITE_LONGBOAT = 372
    ELITE_LONGBOWMAN = 360
    ELITE_MAGYAR_HUSZAR = 472
    ELITE_MAMELUKE = 368
    ELITE_MANGUDAI = 371
    ELITE_ORGAN_GUN = 563
    ELITE_PLUMED_ARCHER = 27
    ELITE_RATTAN_ARCHER = 621
    ELITE_SAMURAI = 366
    ELITE_SHOTEL_WARRIOR = 569
    ELITE_SKIRMISHER = 98
    ELITE_STEPPE_LANCER = 715
    ELITE_TARKAN = 2
    ELITE_TEUTONIC_KNIGHT = 364
    ELITE_THROWING_AXEMAN = 363
    ELITE_TURTLE_SHIP = 448
    ELITE_WAR_ELEPHANT = 367
    ELITE_WAR_WAGON = 450
    ELITE_WOAD_RAIDER = 370
    ENABLE_COWS = 557
    ENABLE_LLAMAS = 556
    ENABLE_SHEEP = 555
    ENABLE_TURKEYS = 558
    ETHIOPIANS = 581
    FAITH = 45
    FARIMBA = 577
    FAST_FIRE_SHIP = 246
    FERVOR = 252
    FEUDAL_AGE = 101
    FIRE_TOWER = 527
    FLETCHING = 199
    FORCED_LEVY = 625
    FORGING = 67
    FORTIFIED_WALL = 194
    FRANKS = 530
    FREE_CARTOGRAPHY = 600
    FUROR_CELTICA = 5
    GALLEON = 35
    GARLAND_WARS = 24
    BONFIRE = 65
    GOLD_MINING = 55
    GOLD_SHAFT_MINING = 182
    GOTHS = 531
    GREAT_WALL = 462
    GREEK_FIRE = 464
    GUARD_TOWER = 140
    GUILDS = 15
    HALBERDIER = 429
    HAND_CANNON = 85
    HAND_CART = 249
    HEATED_SHOT = 380
    HEAVY_CAMEL_RIDER = 236
    HEAVY_CAV_ARCHER = 218
    HEAVY_DEMOLITION_SHIP = 244
    HEAVY_PLOW = 13
    HEAVY_SCORPION = 239
    HERBAL_MEDICINE = 441
    HERESY = 439
    HILL_FORTS_2 = 394
    HILL_FORTS = 691
    HOARDINGS = 379
    HORSE_COLLAR = 14
    HOWDAH = 626
    HUNS = 545
    HUNTING_DOGS = 526
    HUSBANDRY = 39
    HUSSAR = 428
    ILLUMINATION = 233
    IMPERIAL_AGE = 103
    IMPERIAL_CAMEL_RIDER = 521
    IMPERIAL_SKIRMISHER = 655
    INCAS = 549
    INDIANS = 548
    INQUISITION = 492
    IRON_CASTING = 68
    IRONCLAD = 489
    ITALIANS = 547
    JAPANESE = 533
    KAMANDARAN = 488
    KASBAH = 578
    KATAPARUTO = 59
    KEEP = 63
    KHMER = 650
    KOREANS = 546
    LEATHER_ARCHER_ARMOR = 212
    LIGHT_CAVALRY = 254
    LITHUANIANS = 676
    LOGISTICA = 61
    LONG_SWORDSMAN = 207
    LOOM = 22
    MADRASAH = 490
    MAGHRABI_CAMELS = 579
    MAGYARS = 550
    MAHOUTS = 7
    MALAY = 651
    MALIANS = 582
    MAN_AT_ARMS = 222
    MANIPUR_CAVALRY = 627
    MARAUDERS = 483
    MASONRY = 50
    MAYANS = 544
    MONGOLS = 540
    MURDER_HOLES = 322
    NOMADS = 487
    OBSIDIAN_ARROWS = 485
    ONAGER = 257
    ORTHODOXY = 512
    PADDED_ARCHER_ARMOR = 211
    PALADIN = 265
    PANOKSEON = 486
    PAPER_MONEY = 629
    PARTHIAN_TACTICS = 436
    PAVISE = 494
    PERFUSION = 457
    PERSIANS = 536
    PIKEMAN = 197
    PLATE_BARDING_ARMOR = 80
    PLATE_MAIL_ARMOR = 77
    PORTUGUESE = 580
    RECURVE_BOW = 515
    REDEMPTION = 316
    REVETMENTS = 525
    RING_ARCHER_ARMOR = 219
    ROCKETRY = 52
    ROYAL_HEIRS = 574
    SANCTITY = 231
    SAPPERS = 321
    SARACENS = 537
    SCALE_BARDING_ARMOR = 81
    SCALE_MAIL_ARMOR = 74
    SCORPION = 94
    SET_MAXIMUM_POPULATION_NO_HOUSES = 658
    SHATAGNI = 507
    SHINKICHON = 445
    SHIPWRIGHT = 373
    SIEGE_ENGINEERS = 377
    SIEGE_ONAGER = 320
    SIEGE_RAM = 255
    SILK_ARMOR = 687
    SILK_ROAD = 499
    SIPAHI = 491
    SLAVS = 551
    SPANISH = 542
    SPIES_AND_TREASON = 408
    SQUIRES = 215
    STEPPE_HUSBANDRY = 689
    STIRRUPS = 685
    STONE_MINING = 278
    STONE_SHAFT_MINING = 279
    STRONGHOLD = 482
    SULTANS = 506
    SUPPLIES = 716
    SUPREMACY = 440
    TATARS = 674
    TEUTONS = 532
    THALASSOCRACY = 624
    THEOCRACY = 438
    THUMB_RING = 437
    TIGUI = 576
    TIMURID_SIEGECRAFT = 688
    TORSION_ENGINES = 575
    TOWER_SHIELDS = 692
    TOWN_CENTER_SPAWN = 639
    TOWN_PATROL = 280
    TOWN_WATCH = 8
    TRACKING = 90
    TREADMILL_CRANE = 54
    TURKS = 538
    TUSK_SWORDS = 622
    TWO_HANDED_SWORDSMAN = 217
    TWO_MAN_SAW = 221
    VIETNAMESE = 653
    VIETNAMESE_VISION = 665
    VIKINGS = 539
    WAR_GALLEY = 34
    WARWOLF = 461
    WHEELBARROW = 213
    YASAMA = 484
    YEOMEN = 3
    ZEALOTRY = 9


tech_names = bidict({
    2: "elite_tarkan",
    3: "yeomen",
    4: "el_dorado",
    5: "furor_celtica",
    6: "drill",
    7: "mahouts",
    8: "town_watch",
    9: "zealotry",
    10: "artillery",
    11: "crenellations",
    12: "crop_rotation",
    13: "heavy_plow",
    14: "horse_collar",
    15: "guilds",
    16: "anarchy",
    17: "banking",
    19: "cartography",
    21: "atheism",
    22: "loom",
    23: "coinage",
    24: "garland_wars",
    27: "elite_plumed_archer",
    34: "war_galley",
    35: "galleon",
    37: "cannon_galleon",
    39: "husbandry",
    45: "faith",
    47: "chemistry",
    48: "caravan",
    49: "berserkergang",
    50: "masonry",
    51: "architecture",
    52: "rocketry",
    54: "treadmill_crane",
    55: "gold_mining",
    59: "kataparuto",
    60: "elite_conquistador",
    61: "logistica",
    63: "keep",
    64: "bombard_tower",
    65: "bonfire",
    67: "forging",
    68: "iron_casting",
    74: "scale_mail_armor",
    75: "blast_furnace",
    76: "chain_mail_armor",
    77: "plate_mail_armor",
    80: "plate_barding_armor",
    81: "scale_barding_armor",
    82: "chain_barding_armor",
    83: "bearded_axe",
    85: "hand_cannon",
    90: "tracking",
    93: "ballistics",
    94: "scorpion",
    96: "capped_ram",
    98: "elite_skirmisher",
    100: "crossbowman",
    101: "feudal_age",
    102: "castle_age",
    103: "imperial_age",
    104: "dark_age",
    140: "guard_tower",
    182: "gold_shaft_mining",
    188: "bombard_cannon",
    194: "fortified_wall",
    197: "pikeman",
    199: "fletching",
    200: "bodkin_arrow",
    201: "bracer",
    202: "double_bit_axe",
    203: "bow_saw",
    207: "long_swordsman",
    209: "cavalier",
    211: "padded_archer_armor",
    212: "leather_archer_armor",
    213: "wheelbarrow",
    215: "squires",
    217: "two_handed_swordsman",
    218: "heavy_cav_archer",
    219: "ring_archer_armor",
    221: "two_man_saw",
    222: "man_at_arms",
    230: "block_printing",
    231: "sanctity",
    233: "illumination",
    236: "heavy_camel_rider",
    237: "arbalester",
    239: "heavy_scorpion",
    244: "heavy_demolition_ship",
    246: "fast_fire_ship",
    249: "hand_cart",
    252: "fervor",
    254: "light_cavalry",
    255: "siege_ram",
    257: "onager",
    264: "champion",
    265: "paladin",
    278: "stone_mining",
    279: "stone_shaft_mining",
    280: "town_patrol",
    287: "franks_free_farming_1",
    315: "conscription",
    316: "redemption",
    319: "atonement",
    320: "siege_onager",
    321: "sappers",
    322: "murder_holes",
    360: "elite_longbowman",
    361: "elite_cataphract",
    362: "elite_chu_ko_nu",
    363: "elite_throwing_axeman",
    364: "elite_teutonic_knight",
    365: "elite_huskarl",
    366: "elite_samurai",
    367: "elite_war_elephant",
    368: "elite_mameluke",
    369: "elite_janissary",
    370: "elite_woad_raider",
    371: "elite_mangudai",
    372: "elite_longboat",
    373: "shipwright",
    374: "careening",
    375: "dry_dock",
    376: "elite_cannon_galleon",
    377: "siege_engineers",
    379: "hoardings",
    380: "heated_shot",
    384: "eagle_warrior",
    394: "hill_forts_2",
    396: "chinese_50_percent_hp_demos",
    398: "elite_berserk",
    408: "spies_and_treason",
    428: "hussar",
    429: "halberdier",
    432: "elite_jaguar_warrior",
    434: "elite_eagle_warrior",
    435: "bloodlines",
    436: "parthian_tactics",
    437: "thumb_ring",
    438: "theocracy",
    439: "heresy",
    440: "supremacy",
    441: "herbal_medicine",
    445: "shinkichon",
    448: "elite_turtle_ship",
    450: "elite_war_wagon",
    457: "perfusion",
    460: "atlatl",
    461: "warwolf",
    462: "great_wall",
    463: "chieftains",
    464: "greek_fire",
    468: "elite_genoese_crossbowman",
    472: "elite_magyar_huszar",
    481: "elite_elephant_archer",
    482: "stronghold",
    483: "marauders",
    484: "yasama",
    485: "obsidian_arrows",
    486: "panokseon",
    487: "nomads",
    488: "kamandaran",
    489: "ironclad",
    490: "madrasah",
    491: "sipahi",
    492: "inquisition",
    493: "chivalry",
    494: "pavise",
    499: "silk_road",
    504: "elite_boyar",
    506: "sultans",
    507: "shatagni",
    509: "elite_kamayuk",
    512: "orthodoxy",
    513: "druzhina",
    514: "corvinian_army",
    515: "recurve_bow",
    516: "andean_sling",
    517: "couriers",
    521: "imperial_camel_rider",
    525: "revetments",
    526: "hunting_dogs",
    527: "fire_tower",
    529: "britons",
    530: "franks",
    531: "goths",
    532: "teutons",
    533: "japanese",
    534: "chinese",
    535: "byzantines",
    536: "persians",
    537: "saracens",
    538: "turks",
    539: "vikings",
    540: "mongols",
    541: "celts",
    542: "spanish",
    543: "aztecs",
    544: "mayans",
    545: "huns",
    546: "koreans",
    547: "italians",
    548: "indians",
    549: "incas",
    550: "magyars",
    551: "slavs",
    555: "enable_sheep",
    556: "enable_llamas",
    557: "enable_cows",
    558: "enable_turkeys",
    563: "elite_organ_gun",
    565: "elite_camel_archer",
    567: "elite_gbeto",
    569: "elite_shotel_warrior",
    572: "carrack",
    573: "arquebus",
    574: "royal_heirs",
    575: "torsion_engines",
    576: "tigui",
    577: "farimba",
    578: "kasbah",
    579: "maghrabi_camels",
    580: "portuguese",
    581: "ethiopians",
    582: "malians",
    583: "berbers",
    597: "elite_caravel",
    599: "elite_genitour",
    600: "free_cartography",
    602: "arson",
    608: "arrowslits",
    615: "elite_ballista_elephant",
    617: "elite_karambit_warrior",
    619: "elite_arambai",
    621: "elite_rattan_archer",
    622: "tusk_swords",
    623: "double_crossbow",
    624: "thalassocracy",
    625: "forced_levy",
    626: "howdah",
    627: "manipur_cavalry",
    628: "chatras",
    629: "paper_money",
    631: "elite_battle_elephant",
    639: "town_center_spawn",
    650: "khmer",
    651: "malay",
    652: "burmese",
    653: "vietnamese",
    655: "imperial_skirmisher",
    658: "set_maximum_population_no_houses",
    665: "vietnamese_vision",
    673: "bulgarians",
    674: "tatars",
    675: "cumans",
    676: "lithuanians",
    678: "elite_konnik",
    680: "elite_keshik",
    682: "elite_kipchak",
    684: "elite_leitis",
    685: "stirrups",
    686: "bagains",
    687: "silk_armor",
    688: "timurid_siegecraft",
    689: "steppe_husbandry",
    690: "cuman_mercenaries",
    691: "hill_forts",
    692: "tower_shields",
    715: "elite_steppe_lancer",
    716: "supplies",
})



from enum import IntEnum

from bidict import bidict


class Hero(IntEnum):
    LIEF_ERIKSON = 106
    RICHARD_THE_LIONHEART = 160
    THE_BLACK_PRINCE = 161
    FRIAR_TUCK = 163
    SHERIFF_OF_NOTTINGHAM = 164
    CHARLEMAGNE = 165
    ROLAND = 166
    BELISARIUS = 167
    THEODORIC_THE_GOTH = 168
    AETHELFRITH = 169
    SIEGFRIED = 170
    ERIK_THE_RED = 171
    TAMERLANE = 172
    KING_ARTHUR = 173
    LANCELOT = 174
    GAWAIN = 175
    MORDRED = 176
    ARCHBISHOP = 177
    VLAD_DRACULA = 193
    KITABATAKE = 195
    MINAMOTO = 196
    ALEXANDER_NEVSKI = 197
    EL_CID = 198
    ROBIN_HOOD = 200
    VASCO_DA_GAMA = 203
    ALARIC_THE_GOTH = 223
    KING_BELA_IV = 230
    CUAUHTEMOC = 307
    HENRY_THE_LION = 418
    CHARLES_MARTEL = 424
    FRANCISCO_DE_ORELLANA = 425
    HAROLD_HARDRADA = 426
    GONZALO_PIZARRO = 427
    HROLF_THE_GANGER = 428
    FREDERICK_BARBAROSSA = 429
    JOAN_THE_MAID = 430
    WILLIAM_WALLACE = 432
    PRITHVIRAJ = 437
    FRANCESCO_SFORZA = 439
    ATAULF = 453
    JOAN_OF_ARC = 629
    FRANKISH_PALADIN = 632
    SIEUR_DE_METZ = 634
    SIEUR_BERTRAND = 636
    DUKE_D_ALENASONN = 638
    LA_HIRE = 640
    LORD_DE_GRAVILLE = 642
    JEAN_DE_LORRAIN = 644
    CONSTABLE_RICHEMONT = 646
    GUY_JOSSELYNE = 648
    JEAN_BUREAU = 650
    SIR_JOHN_FASTOLF = 652
    REYNALD_DE_CHATILLON = 678
    MASTER_OF_THE_TEMPLAR = 680
    BAD_NEIGHBOR = 682
    GODS_OWN_SLING = 683
    ARCHER_OF_THE_EYES = 686
    SUBOTAI = 698
    HUNTING_WOLF = 700
    KUSHLUK = 702
    TOPA_YUPANQUI = 703
    SHAH = 704
    SABOTEUR = 706
    ORNLU_THE_WOLF = 707
    GODS_OWN_SLING_PACKED = 729
    BAD_NEIGHBOR_PACKED = 730
    GENGHIS_KHAN = 731
    EMPEROR_IN_A_BARREL = 733
    CUSI_YUPANQUI = 749
    ATTILA_THE_HUN = 777
    BLEDA_THE_HUN = 779
    POPE_LEO_I = 781
    SCYTHIAN_WILD_WOMAN = 783
    EL_CID_CAMPEADOR = 824
    KING_SANCHO = 838
    KING_ALFONSO = 840
    IMAM = 842
    ADMIRAL_YI_SUN_SHIN = 844
    NOBUNAGA = 845
    HENRY_V = 847
    WILLIAM_THE_CONQUEROR = 849
    SCYTHIAN_SCOUT = 852
    MUSA_IBN_NUSAYR = 1034
    SUNDJATA = 1035
    TARIQ_IBN_ZIYAD = 1036
    RICHARD_DE_CLARE = 1037
    TRISTAN = 1038
    PRINCESS_YODIT = 1039
    HENRY_II = 1040
    YEKUNA_AMLAK = 1064
    YODIT = 1066
    ITZCOATL = 1067
    MUSTAFA_PASHA = 1068
    PACAL_II = 1069
    BABUR = 1070
    ABRAHA_ELEPHANT = 1071
    GUGLIELMO_EMBRIACO = 1072
    SU_DINGFANG = 1073
    PACHACUTI = 1074
    HUAYNA_CAPAC = 1075
    MIKLOS_TOLDI = 1076
    LITTLE_JOHN = 1077
    ZAWISZA_THE_BLACK = 1078
    SUMANGURU = 1080
    DAGNAJAN = 1106
    GIDAJAN = 1109
    GAJAH_MADA = 1157
    JAYANEGARA = 1158
    RADEN_WIJAYA = 1159
    SUNDA_ROYAL_FIGHTER = 1160
    SURYAVARMAN_I = 1162
    UDAYADITYAVARMAN_I = 1163
    JAYAVIRAVARMAN = 1164
    BAYINNAUNG = 1165
    TABINSHWEHTI = 1166
    LE_LOI = 1178
    LE_LAI = 1180
    LE_TRIEN = 1181
    LUU_NHAN_CHU = 1182
    BUI_BI = 1183
    DINH_LE = 1184
    WANG_TONG = 1185
    ENVOY = 1186
    TOKHTAMYSH_KHAN = 1262
    IVAYLO = 1265
    TSAR_KONSTANTIN = 1266
    KOTYAN_KHAN = 1267
    CUMAN_CHIEF = 1268
    GIRGEN_KHAN = 1269
    URUS_KHAN = 1276
    VYTAUTAS_THE_GREAT = 1281
    IVAYLO_DISMOUNTED = 1290
    SANYOGITA = 1293
    PRITHVI = 1294
    CHAND_BARDAI = 1295
    SALADIN = 1296
    KHOSRAU = 1297
    JARL = 1298
    SAVAR = 1299
    OSMAN = 1303
    MOUNTED_SAMURAI = 1568
    SOSSO_GUARD = 1574
    THE_MIDDLEBROOK = 1631


hero_names = bidict({
    106: "lief_erikson",
    160: "richard_the_lionheart",
    161: "the_black_prince",
    163: "friar_tuck",
    164: "sheriff_of_nottingham",
    165: "charlemagne",
    166: "roland",
    167: "belisarius",
    168: "theodoric_the_goth",
    169: "aethelfrith",
    170: "siegfried",
    171: "erik_the_red",
    172: "tamerlane",
    173: "king_arthur",
    174: "lancelot",
    175: "gawain",
    176: "mordred",
    177: "archbishop",
    193: "vlad_dracula",
    195: "kitabatake",
    196: "minamoto",
    197: "alexander_nevski",
    198: "el_cid",
    200: "robin_hood",
    203: "vasco_da_gama",
    223: "alaric_the_goth",
    230: "king_bela_iv",
    307: "cuauhtemoc",
    418: "henry_the_lion",
    424: "charles_martel",
    425: "francisco_de_orellana",
    426: "harold_hardrada",
    427: "gonzalo_pizarro",
    428: "hrolf_the_ganger",
    429: "frederick_barbarossa",
    430: "joan_the_maid",
    432: "william_wallace",
    437: "prithviraj",
    439: "francesco_sforza",
    453: "ataulf",
    629: "joan_of_arc",
    632: "frankish_paladin",
    634: "sieur_de_metz",
    636: "sieur_bertrand",
    638: "duke_d'alenã§onn",
    640: "la_hire",
    642: "lord_de_graville",
    644: "jean_de_lorrain",
    646: "constable_richemont",
    648: "guy_josselyne",
    650: "jean_bureau",
    652: "sir_john_fastolf",
    678: "reynald_de_chatillon",
    680: "master_of_the_templar",
    682: "bad_neighbor",
    683: "god's_own_sling",
    686: "archer_of_the_eyes",
    698: "subotai",
    700: "hunting_wolf",
    702: "kushluk",
    703: "topa_yupanqui",
    704: "shah",
    706: "saboteur",
    707: "ornlu_the_wolf",
    729: "god's_own_sling_(packed)",
    730: "bad_neighbor_(packed)",
    731: "genghis_khan",
    733: "emperor_in_a_barrel",
    749: "cusi_yupanqui",
    777: "attila_the_hun",
    779: "bleda_the_hun",
    781: "pope_leo_i",
    783: "scythian_wild_woman",
    824: "el_cid_campeador",
    838: "king_sancho",
    840: "king_alfonso",
    842: "imam",
    844: "admiral_yi_sun-shin",
    845: "nobunaga",
    847: "henry_v",
    849: "william_the_conqueror",
    852: "scythian_scout",
    1034: "musa_ibn_nusayr",
    1035: "sundjata",
    1036: "tariq_ibn_ziyad",
    1037: "richard_de_clare",
    1038: "tristan",
    1039: "princess_yodit",
    1040: "henry_ii",
    1064: "yekuna_amlak",
    1066: "yodit",
    1067: "itzcoatl",
    1068: "mustafa_pasha",
    1069: "pacal_ii",
    1070: "babur",
    1071: "abraha_elephant",
    1072: "guglielmo_embriaco",
    1073: "su_dingfang",
    1074: "pachacuti",
    1075: "huayna_capac",
    1076: "miklos_toldi",
    1077: "little_john",
    1078: "zawisza_the_black",
    1080: "sumanguru",
    1106: "dagnajan",
    1109: "gidajan",
    1157: "gajah_mada",
    1158: "jayanegara",
    1159: "raden_wijaya",
    1160: "sunda_royal_fighter",
    1162: "suryavarman_i",
    1163: "udayadityavarman_i",
    1164: "jayaviravarman",
    1165: "bayinnaung",
    1166: "tabinshwehti",
    1178: "le_loi",
    1180: "le_lai",
    1181: "le_trien",
    1182: "luu_nhan_chu",
    1183: "bui_bi",
    1184: "dinh_le",
    1185: "wang_tong",
    1186: "envoy",
    1262: "tokhtamysh_khan",
    1265: "ivaylo",
    1266: "tsar_konstantin",
    1267: "kotyan_khan",
    1268: "cuman_chief",
    1269: "girgen_khan",
    1276: "urus_khan",
    1281: "vytautas_the_great",
    1290: "ivaylo_dismounted",
    1293: "sanyogita",
    1294: "prithvi",
    1295: "chand_bardai",
    1296: "saladin",
    1297: "khosrau",
    1298: "jarl",
    1299: "savar",
    1303: "osman",
    1568: "mounted_samurai",
    1574: "sosso_guard",
    1631: "the_middlebrook",
})


from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces.structs.aoe2_struct import AoE2Struct
from AoE2ScenarioParser.pieces.structs.unit import UnitStruct


class PlayerUnitsStruct(AoE2Struct):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Unit count", DataType("u32"), save_as="unit_count"),
            Retriever("Units", DataType(UnitStruct), set_repeat="{unit_count}")
        ]

        super().__init__("Player Units", retrievers, parser_obj, data)



import collections
import time
import zlib

from AoE2ScenarioParser.helper import generator
from AoE2ScenarioParser.helper import parser
from AoE2ScenarioParser.helper.helper import create_textual_hex, SimpleLogger
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object_manager import AoE2ObjectManager
from AoE2ScenarioParser.pieces.background_image import BackgroundImagePiece
from AoE2ScenarioParser.pieces.cinematics import CinematicsPiece
from AoE2ScenarioParser.pieces.data_header import DataHeaderPiece
from AoE2ScenarioParser.pieces.diplomacy import DiplomacyPiece
from AoE2ScenarioParser.pieces.file_header import FileHeaderPiece
from AoE2ScenarioParser.pieces.global_victory import GlobalVictoryPiece
from AoE2ScenarioParser.pieces.map import MapPiece
from AoE2ScenarioParser.pieces.messages import MessagesPiece
from AoE2ScenarioParser.pieces.options import OptionsPiece
from AoE2ScenarioParser.pieces.player_data_two import PlayerDataTwoPiece
from AoE2ScenarioParser.pieces.triggers import TriggerPiece
from AoE2ScenarioParser.pieces.units import UnitsPiece


class AoE2Scenario:
    def __init__(self, filename, log_reading=True, log_parsing=False):
        print("\nPreparing & Loading file: '" + filename + "'...")

        scenario = open(filename, "rb")
        self._file = scenario.read()
        scenario.seek(0)  # Reset file cursor to 0

        self._file_header = scenario.read(self._compute_header_length())
        self._file_data = zlib.decompress(scenario.read(), -zlib.MAX_WBITS)

        scenario.close()
        # self._debug_write_from_source("../PreParseErrorFile", "hd", write_in_bytes=False)
        # exit()

        print("File prepared and loaded.")

        self.parser = parser.Parser()
        self._read_file(log_reading=log_reading)
        self.object_manager = AoE2ObjectManager(self._parsed_header, self._parsed_data, log_parsing=log_parsing)

    def _read_file(self, log_reading):
        lgr = SimpleLogger(should_log=log_reading)
        lgr.print("\nFile reading started...")

        self._parsed_header = collections.OrderedDict()
        self._parsed_data = collections.OrderedDict()
        header_generator = self._create_header_generator(1)
        data_generator = self._create_data_generator(1)

        current_piece = ""
        try:
            for piece_object in _header_structure:
                piece = piece_object(self.parser)
                piece_name = type(piece).__name__
                self._parsed_header[piece_name] = piece
                current_piece = piece_name

                lgr.print("\tReading " + piece_name + "...")
                piece.set_data_from_generator(header_generator)
                lgr.print("\tReading " + piece_name + " finished successfully.")

            for piece_object in _file_structure:
                piece = piece_object(self.parser)
                piece_name = type(piece).__name__
                self._parsed_data[piece_name] = piece
                current_piece = piece_name

                lgr.print("\tReading " + piece_name + "...")
                piece.set_data_from_generator(data_generator)
                lgr.print("\tReading " + piece_name + " finished successfully.")
        except StopIteration as e:
            print(f"\n[StopIteration] [EXIT] AoE2Scenario._read_file: \n\tPiece: {current_piece}\n")
            print("Writing ErrorFile...")
            self._debug_byte_structure_to_file("../ErrorFile", log_debug_write=False)
            time.sleep(1)
            print("ErrorFile written. \n\n\n ------------------------ STACK TRACE ------------------------\n\n")
            time.sleep(1)
            raise StopIteration(e)

        suffix = b''
        try:
            while True:
                suffix += data_generator.__next__()
        except StopIteration:
            # End of file reached
            pass
        finally:
            if len(suffix) > 0:
                # print("Found file suffix! Length: " + str(len(suffix)) + ". Suffix content: '" + str(suffix) + "'.")
                pass
            self._suffix = suffix

        lgr.print("File reading finished successfully.")

    def write_to_file(self, filename, log_writing=True, log_reconstructing=False):
        self._write_from_structure(filename, log_writing=log_writing, log_reconstructing=log_reconstructing)

    def _write_from_structure(self, filename, write_in_bytes=True, compress=True,
                              log_writing=True, log_reconstructing=False):
        if hasattr(self, 'object_manager'):
            self.object_manager.reconstruct(log_reconstructing=log_reconstructing)
        lgr = SimpleLogger(should_log=log_writing)
        lgr.print("\nFile writing from structure started...")

        byte_header = b''
        byte_data = b''

        for key in self._parsed_header:
            lgr.print("\twriting " + key + "...")
            for retriever in self._parsed_header[key].retrievers:
                byte_header += parser.retriever_to_bytes(retriever)
            lgr.print("\twriting " + key + " finished successfully.")

        for key in self._parsed_data:
            lgr.print("\twriting " + key + "...")
            for retriever in self._parsed_data[key].retrievers:
                try:
                    byte_data += parser.retriever_to_bytes(retriever)
                except AttributeError as e:
                    print("AttributeError occurred while writing '" + key + "' > '" + retriever.name + "'")
                    print("\n\n\nAn error occurred. Writing failed.")
                    raise e
            lgr.print("\twriting " + key + " finished successfully.")

        file = open(filename, "wb" if write_in_bytes else "w")
        file.write(byte_header if write_in_bytes else create_textual_hex(byte_header.hex()))

        if compress:
            # https://stackoverflow.com/questions/3122145/zlib-error-error-3-while-decompressing-incorrect-header-check/22310760#22310760
            deflate_obj = zlib.compressobj(9, zlib.DEFLATED, -zlib.MAX_WBITS)
            compressed = deflate_obj.compress(byte_data + self._suffix) + deflate_obj.flush()
            file.write(compressed if write_in_bytes else create_textual_hex(compressed.hex()))
        else:
            file.write(byte_data if write_in_bytes else create_textual_hex(byte_data.hex()))

        file.close()
        lgr.print("File writing finished successfully.")

    def _create_header_generator(self, chunk_size):
        return generator.create_generator(self._file_header, chunk_size)

    def _create_data_generator(self, chunk_size):
        return generator.create_generator(self._file_data, chunk_size)

    def _create_file_generator(self, chunk_size):
        return generator.create_generator(self._file, chunk_size)

    def _compute_header_length(self):
        return parser.calculate_length(
            self._create_file_generator(1),
            FileHeaderPiece(parser.Parser()).retrievers
        )

    """ #############################################
    ################ Debug functions ################
    ############################################# """

    def _debug_log_all_data_keys(self):
        """ Used for debugging. """
        print("FileHeader:")
        print(self._parsed_header.keys())
        print("DataHeader:")
        print("odict_keys([")
        for x in self._parsed_data.keys():
            print("\t'" + x + "'")
        print("])")

    def _debug_log_effect_dataset(self):
        """ Used for debugging - Only reads One Trigger. """
        trigger_data = find_retriever(self._parsed_data['TriggerPiece'].retrievers, "Trigger data").data
        effects = parser.listify(find_retriever(trigger_data.retrievers, "Effect data").data)

        for effect in effects:
            for retriever in effect.retrievers:
                # if retriever.data != -1 and \
                #         retriever.data != [] and \
                #         retriever.data != "" and \
                #         retriever.data != " " and \
                #         retriever.name != "static_value_46":
                if retriever.name != "static_value_46":
                    if retriever.name == "effect_type":
                        print("},\n" + str(retriever.data) + ": {")
                    print("\t\"" + retriever.name + "\": " +
                          (str(retriever.data) if type(retriever.data) is not str else "\"" + retriever.data + "\"")
                          + ",")
        print("}\n")

    def _debug_log_condition_dataset(self):
        """ Used for debugging - Only reads One Trigger. """
        trigger_data = find_retriever(self._parsed_data['TriggerPiece'].retrievers, "Trigger data").data
        conditions = parser.listify(find_retriever(trigger_data.retrievers, "Condition data").data)

        for condition in conditions:
            for retriever in condition.retrievers:
                # if retriever.data != -1 and \
                #         retriever.data != [] and \
                #         retriever.data != "" and \
                #         retriever.data != " " and \
                #         retriever.name != "static_value_21":
                if retriever.name != "static_value_21":
                    if retriever.name == "condition_type":
                        print("},\n" + str(retriever.data) + ": {")
                    print("\t\"" + retriever.name + "\": " +
                          (str(retriever.data) if type(retriever.data) is not str else "\"" + retriever.data + "\"")
                          + ",")
        print("}\n")

    def _debug_write_from_source(self, filename, datatype, write_in_bytes=True):
        """This function is used as a test debugging writing. It writes parts of the read file to the filesystem."""
        print("File writing from source started with attributes " + datatype + "...")
        file = open(filename, "wb" if write_in_bytes else "w")
        for t in datatype:
            if t == "f":
                file.write(self._file if write_in_bytes else create_textual_hex(self._file.hex()))
            elif t == "h":
                file.write(self._file_header if write_in_bytes else create_textual_hex(self._file_header.hex()))
            elif t == "d":
                file.write(self._file_data if write_in_bytes else create_textual_hex(self._file_data.hex()))
        file.close()
        print("File writing finished successfully.")

    def _debug_byte_structure_to_file(self, filename, log_debug_write=True):
        """ Used for debugging - Writes structure from read file to the filesystem in a easily readable manner. """
        lgr = SimpleLogger(should_log=log_debug_write)

        lgr.print("\nWriting structure to file...")
        with open(filename, 'w') as output_file:
            result = ""
            for key in self._parsed_header:
                lgr.print("\tWriting " + key + "...")
                result += self._parsed_header[key].get_byte_structure_as_string()
                lgr.print("\tWriting " + key + " finished successfully.")
            for key in self._parsed_data:
                lgr.print("\tWriting " + key + "...")
                result += self._parsed_data[key].get_byte_structure_as_string()
                lgr.print("\tWriting " + key + " finished successfully.")

            output_file.write(result)
            output_file.close()
        lgr.print("Writing structure to file finished successfully.")


_header_structure = [
    FileHeaderPiece
]
_file_structure = [
    DataHeaderPiece,
    MessagesPiece,
    CinematicsPiece,
    BackgroundImagePiece,
    PlayerDataTwoPiece,
    GlobalVictoryPiece,
    DiplomacyPiece,
    OptionsPiece,
    MapPiece,
    UnitsPiece,
    TriggerPiece,
]



from enum import IntEnum


class Terrain(IntEnum):
    BEACH = 2
    BEACH_NON_NAVIGABLE = 79
    BEACH_NON_NAVIGABLE_WET_GRAVEL = 81
    BEACH_NON_NAVIGABLE_WET_ROCK = 82
    BEACH_NON_NAVIGABLE_WET_SAND = 80
    BEACH_ICE = 37
    BEACH_VEGETATION = 52
    BEACH_WET = 107
    BEACH_WET_GRAVEL = 108
    BEACH_WET_ROCK = 109
    BEACH_WHITE = 53
    BEACH_WHITE_VEGETATION = 51
    BLACK = 47
    DESERT_CRACKED = 45
    DESERT_QUICKSAND = 46
    DESERT_SAND = 14
    DIRT_1 = 6
    DIRT_2 = 11
    DIRT_3 = 3
    DIRT_4 = 42
    DIRT_MUD = 76
    DIRT_SAVANNAH = 41
    FARM = 7
    FARM_0 = 29
    FARM_33 = 30
    FARM_67 = 31
    FARM_DEAD = 8
    FOREST_ACACIA = 50
    FOREST_AUTUMN = 104
    FOREST_AUTUMN_SNOW = 105
    FOREST_BAMBOO = 18
    FOREST_BAOBAB = 49
    FOREST_BUSH = 89
    FOREST_DEAD = 106
    FOREST_DRAGON_TREE = 48
    FOREST_JUNGLE = 17
    FOREST_MAGROVE = 55
    FOREST_MEDITERRANEAN = 88
    FOREST_OAK = 10
    FOREST_OAK_BUSH = 20
    FOREST_PALM_DESERT = 13
    FOREST_PINE = 19
    FOREST_PINE_SNOW = 21
    FOREST_RAINFOREST = 56
    FOREST_REEDS = 92
    FOREST_REEDS_BEACH = 91
    FOREST_REEDS_SHALLOWS = 90
    GRASS_1 = 0
    GRASS_2 = 12
    GRASS_3 = 9
    GRASS_DRY = 100
    GRASS_FOUNDATION = 27
    GRASS_JUNGLE = 60
    GRASS_JUNGLE_RAINFOREST = 83
    GRASS_OTHER = 16
    GRAVEL_DEFAULT = 70
    GRAVEL_DESERT = 102
    ICE = 35
    ICE_NAVIGABLE = 26
    RICE_FARM = 63
    RICE_FARM_0 = 65
    RICE_FARM_33 = 66
    RICE_FARM_67 = 67
    RICE_FARM_DEAD = 64
    ROAD = 24
    ROAD_BROKEN = 25
    ROAD_FUNGUS = 75
    ROAD_GRAVEL = 78
    ROCK_1 = 40
    SHALLOWS = 4
    SHALLOWS_AZURE = 59
    SHALLOWS_MANGROVE = 54
    SNOW = 32
    SNOW_FOUNDATION = 36
    SNOW_LIGHT = 73
    SNOW_STRONG = 74
    SWAMP_BOGLAND = 101
    UNDERBUSH = 5
    UNDERBUSH_JUNGLE = 77
    UNDERBUSH_LEAVES = 71
    UNDERBUSH_SNOW = 72
    WATER_2D_BRIDGE = 28
    WATER_2D_SHORELESS = 15
    WATER_AZURE = 58
    WATER_BROWN = 96
    WATER_DEEP = 22
    WATER_DEEP_OCEAN = 57
    WATER_GREEN = 95
    WATER_MEDIUM = 23
    WATER_SHALLOW = 1



from __future__ import annotations

from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.aoe2_object import AoE2Object
from AoE2ScenarioParser.pieces.structs.variable import VariableStruct


class VariableObject(AoE2Object):
    def __init__(self,
                 variable_id,
                 name
                 ):
        self.variable_id = variable_id
        self.name = name

        super().__init__()

    @staticmethod
    def _parse_object(parsed_data, **kwargs) -> VariableObject:
        # Expected {variable=VariableStruct}
        variable = kwargs['variable']

        return VariableObject(
            variable_id=find_retriever(variable.retrievers, "variable_id").data,
            name=find_retriever(variable.retrievers, "name").data
        )

    @staticmethod
    def _reconstruct_object(parsed_header, parsed_data, objects, **kwargs) -> None:
        # Expected {variable=VariableObject, variables=variables_list}
        variable_obj = kwargs['variable']
        variables_list = kwargs['variables']

        variables_list.append(VariableStruct(
            data=[variable_obj.variable_id, variable_obj.name]
        ))



from enum import IntEnum

from AoE2ScenarioParser.datasets.buildings import Building, GaiaBuilding
from AoE2ScenarioParser.datasets.heroes import Hero
from AoE2ScenarioParser.datasets.units import Unit, GaiaUnit

""" =============================================================
========================= HEX FUNCTIONS =========================
=============================================================="""


def update_order_array(order_array, supposed_length):
    for i in range(0, supposed_length):
        if i not in order_array:
            order_array.append(i)


def create_textual_hex(string, space_distance=2, enter_distance=48):
    """Please note that the 'enter_distance' parameter is including the - to be added - spaces. If you calculated it
    without the spaces, please multiply the number by: `block size incl space / block size excl space`"""
    return insert_char(insert_char(string, " ", space_distance), "\n", enter_distance)


# Credits: gurney alex @ https://stackoverflow.com/a/2657733/7230293
def insert_char(string, char, every=64):
    return char.join(string[i:i + every] for i in range(0, len(string), every))


""" =============================================================
======================= STRING MODIFIERS ========================
=============================================================="""


def add_str_trail(string):
    return (string + ("\x00" if string[-1] != "\x00" else "")) if len(string) > 0 else string


def del_str_trail(string):
    return string.replace('\x00', "")


def add_prefix_chars(string, char, length):
    if len(string) > length:
        return string
    else:
        return char * (length - len(string)) + string


def add_suffix_chars(string, char, length):
    if len(string) > length:
        return string
    else:
        return string + char * (length - len(string))


""" =============================================================
======================== PRETTY PRINTERS ========================
=============================================================="""


def pretty_print_list(plist):
    return_string = "[\n"
    for x in plist:
        return_string += "\t" + str(x)
    return return_string + "]\n"


def pretty_print_name(name: str) -> str:
    """
    Returns a pretty-printed version of the name string.
    Replaces all underscores with spaces and capitalizes the first letter
    of each word.
    For example, elite_chu_ko_nu -> Elite Chu Ko Nu.

    :Author:
        T-West (https://github.com/twestura/)
    """
    return ' '.join(s[0].upper() + s[1:] for s in name.split('_'))


""" =============================================================
============================= OTHER =============================
=============================================================="""


def get_enum_from_unit_const(const: int) -> IntEnum:
    """
    Returns an Enum corresponding with the given Const.

    Arguments:
        const: The constant representing a unit
    """
    if any(item == const for item in Unit):
        return Unit(const)
    if any(item.value == const for item in Building):
        return Building(const)
    if any(item.value == const for item in Hero):
        return Hero(const)
    if any(item == const for item in GaiaUnit):
        return GaiaUnit(const)
    if any(item == const for item in GaiaBuilding):
        return GaiaBuilding(const)


class SimpleLogger:
    def __init__(self, should_log):
        self.should_log = should_log

    def print(self, string):
        if self.should_log:
            print(string)

    def __repr__(self):
        return f"SimpleLogger[should_log: {self.should_log}]"


class Tile:
    def __init__(self, x: int, y: int):
        self.x: float = x
        self.y: float = y

    def __repr__(self):
        return f"Tile[x: {self.x}, y: {self.y}]"

    @property
    def x(self) -> float:
        return self._x

    @x.setter
    def x(self, val: int):
        self._x = int(val) + .5

    @property
    def y(self) -> float:
        return self._y

    @y.setter
    def y(self, val: int):
        self._y = int(val) + .5

    @property
    def x1(self) -> int:
        return int(self.x - .5)

    @property
    def y1(self) -> int:
        return int(self.y - .5)

    @property
    def x2(self) -> int:
        return int(self.x + .5)

    @property
    def y2(self) -> int:
        return int(self.y + .5)



from __future__ import annotations

from typing import List

from AoE2ScenarioParser.helper import generator
from AoE2ScenarioParser.helper.helper import SimpleLogger
from AoE2ScenarioParser.helper.retriever import find_retriever
from AoE2ScenarioParser.objects.data_header_obj import DataHeaderObject
from AoE2ScenarioParser.objects.diplomacy_obj import DiplomacyObject
from AoE2ScenarioParser.objects.file_header_obj import FileHeaderObject
from AoE2ScenarioParser.objects.map_obj import MapObject
from AoE2ScenarioParser.objects.messages_obj import MessagesObject
from AoE2ScenarioParser.objects.options_obj import OptionsObject
from AoE2ScenarioParser.objects.player_object import PlayerObject
from AoE2ScenarioParser.objects.terrain_obj import TerrainObject
from AoE2ScenarioParser.objects.triggers_obj import TriggersObject
from AoE2ScenarioParser.objects.units_obj import UnitsObject


class AoE2ObjectManager:
    def __init__(self, parser_header, parsed_data, log_parsing=True):
        lgr = SimpleLogger(should_log=log_parsing)
        lgr.print("\nParsing pieces and structs to objects...")
        self.parser_header = parser_header
        self.parsed_data = parsed_data
        self._objects = {}
        self._finished_new_structure = {
            "UnitsObject": UnitsObject,
            "TriggersObject": TriggersObject,
        }

        # self._objects = {
        #     # "FileHeaderObject": self._parse_file_header_object(),
        #     # "DataHeaderObject": self._parse_data_header_object(),
        #     # "PlayerObject": self._parse_player_object(),
        #     # "MessagesObject": self._parse_messages_object(),
        #     # "DiplomacyObject": self._parse_diplomacy_object(),
        #     # "OptionsObject": self._parse_options_object(),
        #     # "MapObject": self._parse_map_object(),
        #     "UnitsObject": UnitsObject.parse_object(self.parsed_data),
        #     "TriggersObject": TriggersObject.parse_object(self.parsed_data)
        # }

        for key in self._finished_new_structure.keys():
            lgr.print("\tParsing " + key + "...")
            self._objects[key] = self._finished_new_structure[key]._parse_object(self.parsed_data)
            lgr.print("\tParsing " + key + " finished successfully.")

        lgr.print("Parsing pieces and structs to objects finished successfully.")

    @property
    def trigger_manager(self) -> TriggersObject:
        return self._objects['TriggersObject']

    @property
    def unit_manager(self) -> UnitsObject:
        return self._objects['UnitsObject']

    def reconstruct(self, log_reconstructing=False):
        lgr = SimpleLogger(should_log=log_reconstructing)
        lgr.print("\nReconstructing pieces and structs from objects...")

        for key in self._finished_new_structure.keys():
            lgr.print("\tReconstructing " + key + "...")
            self._objects[key]._reconstruct_object(self.parser_header, self.parsed_data, self._objects)
            lgr.print("\tReconstructing " + key + " finished successfully.")

        lgr.print("Reconstruction finished successfully.")

    # ################################################################################################ #
    #                           Todo: Move these functions to their objects.
    # ################################################################################################ #

    def _parse_map_object(self):
        object_piece = self.parsed_data['MapPiece']
        map_width = find_retriever(object_piece.retrievers, "Map Width").data
        map_height = find_retriever(object_piece.retrievers, "Map Height").data
        terrain_list = find_retriever(object_piece.retrievers, "Terrain data").data
        # AoE2 in Game map: Left to top = X. Left to bottom = Y. Tiny map top = [X:199,Y:0]
        terrain_2d = []
        """
        Debating to support only Py3.6+. Until then, this'll be a comment.
        >> terrain_2d: List[List[TerrainObject]] = []
        """

        for i in range(0, map_width * map_height):
            to = TerrainObject(
                terrain_id=find_retriever(terrain_list[i].retrievers, "Terrain ID").data,
                elevation=find_retriever(terrain_list[i].retrievers, "Elevation").data
            )
            map_x = i % map_width
            try:
                terrain_2d[map_x].append(to)
            except IndexError:
                if len(terrain_2d) <= map_x:
                    terrain_2d.append(list())
                terrain_2d[map_x].append(to)

        return MapObject(
            map_color_mood=find_retriever(object_piece.retrievers, "Map color mood").data,
            collide_and_correct=find_retriever(object_piece.retrievers, "Collide and Correcting").data,
            map_width=map_width,
            map_height=map_height,
            terrain=terrain_2d,
        )

    def _parse_options_object(self):
        object_piece = self.parsed_data['OptionsPiece']
        # ppnd: Per Player Number of Disabled
        ppnd_techs = find_retriever(object_piece.retrievers, "Per player number of disabled techs").data
        ppnd_units = find_retriever(object_piece.retrievers, "Per player number of disabled units").data
        ppnd_buildings = find_retriever(object_piece.retrievers, "Per player number of disabled buildings").data
        disabled_techs = generator.create_generator(
            find_retriever(object_piece.retrievers, "Disabled technology IDs in player order").data, 1
        )
        disabled_units = generator.create_generator(
            find_retriever(object_piece.retrievers, "Disabled unit IDs in player order").data, 1
        )
        disabled_buildings = generator.create_generator(
            find_retriever(object_piece.retrievers, "Disabled building IDs in player order").data, 1
        )

        disables = list()
        for player_id in range(0, 8):  # 0-7 Players
            nd_techs = ppnd_techs[player_id]
            nd_units = ppnd_units[player_id]
            nd_buildings = ppnd_buildings[player_id]
            player_disabled_techs = generator.repeat_generator(
                disabled_techs, nd_techs, return_bytes=False)
            player_disabled_units = generator.repeat_generator(
                disabled_units, nd_units, return_bytes=False)
            player_disabled_buildings = generator.repeat_generator(
                disabled_buildings, nd_buildings, return_bytes=False)

            disables.append({
                'techs': player_disabled_techs,
                'units': player_disabled_units,
                'buildings': player_disabled_buildings,
            })

        return OptionsObject(
            disables,
            find_retriever(object_piece.retrievers, "All techs").data
        )

    def _parse_diplomacy_object(self):
        object_piece = self.parsed_data['DiplomacyPiece']
        diplomacy = find_retriever(object_piece.retrievers, "Per-player diplomacy").data

        diplomacies = []
        for player_id in range(0, 8):  # 0-7 Players
            diplomacies.append(find_retriever(diplomacy[player_id].retrievers, "Stance with each player").data)

        return DiplomacyObject(
            player_stances=diplomacies
        )

    def _parse_messages_object(self):
        object_piece = self.parsed_data['MessagesPiece']
        retrievers = object_piece.retrievers

        return MessagesObject(
            instructions=find_retriever(retrievers, "Instructions").data,
            hints=find_retriever(retrievers, "Hints").data,
            victory=find_retriever(retrievers, "Victory").data,
            loss=find_retriever(retrievers, "Loss").data,
            history=find_retriever(retrievers, "History").data,
            scouts=find_retriever(retrievers, "Scouts").data,
            ascii_instructions=find_retriever(retrievers, "ASCII Instructions").data,
            ascii_hints=find_retriever(retrievers, "ASCII Hints").data,
            ascii_victory=find_retriever(retrievers, "ASCII Victory").data,
            ascii_loss=find_retriever(retrievers, "ASCII Loss").data,
            ascii_history=find_retriever(retrievers, "ASCII History").data,
            ascii_scouts=find_retriever(retrievers, "ASCII Scouts").data,
        )

    def _parse_player_object(self):
        players = []

        data_header_piece = self.parsed_data['DataHeaderPiece']
        unit_piece = self.parsed_data['UnitsPiece']
        options_piece = self.parsed_data['OptionsPiece']
        starting_ages = find_retriever(options_piece.retrievers, "Per player starting age").data

        # Player Data
        player_data_one = find_retriever(data_header_piece.retrievers, "Player data#1").data  # 0-7 Players & 8 Gaia
        player_data_two = self.parsed_data['PlayerDataTwoPiece']  # 0-7 Players & 8 Gaia
        resources = find_retriever(player_data_two.retrievers, "Resources").data
        # player_data_three = find_retriever(unit_piece.retrievers, "Player data #3").data  # 0-7 Players
        player_data_four = find_retriever(unit_piece.retrievers, "Player data #4").data  # 0-7 Players

        for player_id in range(0, 9):  # 0-7 Players & 8 Gaia:
            try:  # If gaia isn't saved. (PlayerDataThree and PlayerDataFour)
                pop_limit = find_retriever(player_data_four[player_id].retrievers, "Population limit").data
            except IndexError:
                pop_limit = -1

            players.append(PlayerObject(
                player_number=player_id,
                active=find_retriever(player_data_one[player_id].retrievers, "Active").data,
                human=find_retriever(player_data_one[player_id].retrievers, "Human").data,
                civilization=find_retriever(player_data_one[player_id].retrievers, "Civilization").data,
                gold=find_retriever(resources[player_id].retrievers, "Gold").data,
                wood=find_retriever(resources[player_id].retrievers, "Wood").data,
                food=find_retriever(resources[player_id].retrievers, "Food").data,
                stone=find_retriever(resources[player_id].retrievers, "Stone").data,
                color=find_retriever(resources[player_id].retrievers, "Player color").data,
                starting_age=starting_ages[player_id],
                pop_limit=pop_limit
            ))

        return players

    def _parse_data_header_object(self):
        object_piece = self.parsed_data['DataHeaderPiece']
        retrievers = object_piece.retrievers

        return DataHeaderObject(
            version=find_retriever(retrievers, "Version").data,
            filename=find_retriever(retrievers, "Filename").data
        )

    def _parse_file_header_object(self):
        object_piece = self.parser_header['FileHeaderPiece']
        retrievers = object_piece.retrievers

        return FileHeaderObject(
            version=find_retriever(retrievers, "Version").data,
            timestamp=find_retriever(retrievers, "Timestamp of last save").data,
            instructions=find_retriever(retrievers, "Scenario instructions").data,
            player_count=find_retriever(retrievers, "Player count").data,
            creator_name=find_retriever(retrievers, "Creator name").data,
        )



from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.pieces.structs.player_diplomacy import PlayerDiplomacyStruct


class DiplomacyPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Per-player diplomacy", DataType(PlayerDiplomacyStruct, repeat=16)),
            Retriever("Individual Victories", DataType("60", repeat=16*12)),  # 12 Conditions per (16) Player(s).
            Retriever("Separator", DataType("u32")),
            Retriever("Per-player allied victory", DataType("u32", repeat=16)),  # Ignored -> PlayerDataThree
            Retriever("Unknown", DataType("4")),
        ]

        super().__init__("Diplomacy", retrievers, parser_obj, data=data)



# ------------------- Creating BiDict -----------------------
# from bidict import bidict
#
# from AoE2ScenarioParser.datasets import buildings
#
#
# def get_unit_id_by_string(tech):
#     try:
#         return eval(tech)
#     except NameError:
#         return None
#
#
# units_dict = bidict()
# for x in buildings.__dict__.keys():
#     if '__' not in x and x != 'get_building_id_by_string':
#         units_dict.put(x, buildings.get_building_id_by_string(x))
#
# print("building_names = bidict({")
# for key, value in units_dict.inverse.items():
#     print(" " * 4 + str(key) + ": \"" + str(value) + "\",")
# print("})")

# ------------------- Find Difference ----------------------
# Add IDs to be checked
# a = []
#
# b = []
# for x in a:
#     try:
#         ENUM_HERE(x)
#     except ValueError:
#         b.append(x)
#
# print(b)



class DataType:
    """ A class to identify what data you want to retrieve. This class has two parameters which are very useful.
        var:
            - Almost always a string representation of the data you want to retrieve. This project uses the types from:
                http://dderevjanik.github.io/agescx/formatscx/#format
                s: Signed integer       All integers (signed and unsigned) are parsed as little endian
                u: Unsigned integer     All integers (signed and unsigned) are parsed as little endian
                f: float
                c: Character string of fixed length
                (Empty): Is interpreted as regular byte data. In this project the '' is converted to 'data'
                str: Character string of variable length.
                    This type will read the number in bits given and parse it as an int. The number retrieved from it
                    will be the amount of bytes read as a character string.
            - Another option for the var parameter is to give a Struct (not Python struct) subclass as var. This will
            parse all DataType values in the Struct subclass. This can be handy for when blocks of data are repeated.
            - Per data type you can save how large the value is bit/byte wise. While this may be confusing not all
            values are in bit format. Some are in byte format.
                Bit format: [s, u, f, str]
                Byte format: [c, data]
            - To define a length to your data type, write the datatype with the length behind it (no whitespaces)
                Example 1:  s16             > A signed integer of 16 bits.
                Example 2:  f32             > A 32 bit floating point number.
                Example 3:  c4              > A 32 bit (4 bytes) character string.
                Example 4:  str16           > A 16 bit integer will be parsed (n). Now n bytes will be read as character
                                            string.
                Example 5:  TerrainStruct   > The TerrainStruct will be instantiated and DataTypes from that struct will
                                            be loaded in it's place.
        repeat:
            The amount of times the above datatype needs to be repeated
    """
    def __init__(self, var="0", repeat=1):
        self.var = var
        self.repeat = repeat

    def to_simple_string(self):
        return str(self.repeat) + " * " + (self.var if type(self.var) is str else self.var.__name__)

    def _to_string(self):
        return "[DataType] " + self.to_simple_string()

    def __repr__(self):
        return self._to_string()



from enum import IntEnum

from bidict import bidict


class Condition(IntEnum):
    NONE = 0
    """
    Attributes for the **none**  condition are:\n
    ... It's called none... What parameters do **you** think it has?! ❤ 
    """
    BRING_OBJECT_TO_AREA = 1
    """
    Attributes for the **bring_object_to_area** condition are:\n
    - unit_object  
    - area_1_x
    - area_1_y
    - area_2_x
    - area_2_y
    - inverted
    """
    BRING_OBJECT_TO_OBJECT = 2
    """
    Attributes for the **bring_object_to_object** condition are:\n
    - unit_object
    - next_object
    - inverted
    """
    OWN_OBJECTS = 3
    """
    Attributes for the **own_objects** condition are:\n
    - amount_or_quantity
    - object_list
    - player
    - object_group
    - object_type
    """
    OWH_FEWER_OBJECTS = 4
    """
    Attributes for the **owh_fewer_objects** condition are:\n
    - amount_or_quantity
    - object_list
    - player
    - area_1_x
    - area_1_y
    - area_2_x
    - area_2_y
    - object_group
    - object_type
    """
    OBJECT_IN_AREA = 5
    """
    Attributes for the **object_in_area** condition are:\n
    - amount_or_quantity
    - object_list
    - player
    - area_1_x
    - area_1_y
    - area_2_x
    - area_2_y
    - object_group
    - object_type
    - inverted
    """
    DESTROY_OBJECT = 6
    """
    Attributes for the **destroy_object** condition are:\n
    - unit_object
    - inverted
    """
    CAPTURE_OBJECT = 7
    """
    Attributes for the **capture_object** condition are:\n
    - unit_object
    - player
    - inverted
    """
    ACCUMULATE_ATTRIBUTE = 8
    """
    Attributes for the **accumulate_attribute** condition are:\n
    - amount_or_quantity
    - resource_type_or_tribute_list
    - player
    - inverted
    """
    RESEARCH_TECHNOLOGY = 9
    """
    Attributes for the **research_technology** condition are:\n
    - player
    - technology
    - inverted
    """
    TIMER = 10
    """
    Attributes for the **timer** condition are:\n
    - timer
    - inverted
    """
    OBJECT_SELECTED = 11
    """
    Attributes for the **object_selected** condition are:\n
    - unit_object
    - inverted
    """
    AI_SIGNAL = 12
    """
    Attributes for the **ai_signal** condition are:\n
    - ai_signal
    - inverted
    """
    PLAYER_DEFEATED = 13
    """
    Attributes for the **player_defeated** condition are:\n
    - player
    - inverted
    """
    OBJECT_HAS_TARGET = 14
    """
    Attributes for the **object_has_target** condition are:\n
    - unit_object
    - next_object
    - object_list
    - object_group
    - object_type
    - inverted
    """
    OBJECT_VISIBLE = 15
    """
    Attributes for the **object_visible** condition are:\n
    - unit_object
    """
    OBJECT_NOT_VISIBLE = 16
    """
    Attributes for the **object_not_visible** condition are:\n
    - unit_object
    """
    RESEARCHING_TECH = 17
    """
    Attributes for the **researching_tech** condition are:\n
    - player
    - technology
    - inverted
    """
    UNITS_GARRISONED = 18
    """
    Attributes for the **units_garrisoned** condition are:\n
    - amount_or_quantity
    - unit_object
    - inverted
    """
    DIFFICULTY_LEVEL = 19
    """
    Attributes for the **units_garrisoned** condition are:\n
    - condition_type
    - amount_or_quantity
    - inverted
    """
    CHANCE = 20
    """
    Attributes for the **chance** condition are:\n
    - amount_or_quantity
    """
    TECHNOLOGY_STATE = 21
    """
    Attributes for the **technology_state** condition are:\n
    - amount_or_quantity
    - player
    - technology
    - inverted
    """
    VARIABLE_VALUE = 22
    """
    Attributes for the **variable_value** condition are:\n
    - amount_or_quantity
    - inverted
    - variable
    - comparison
    """
    OBJECT_HP = 23
    """
    Attributes for the **object_hp** condition are:\n
    - amount_or_quantity
    - unit_object
    - inverted
    - comparison
    """
    DIPLOMACY_STATE = 24
    """
    Attributes for the **diplomacy_state** condition are:\n
    - amount_or_quantity
    - player
    - inverted
    - target_player
    """


condition_names = bidict({
    0: "none",
    1: "bring_object_to_area",
    2: "bring_object_to_object",
    3: "own_objects",
    4: "own_fewer_objects",
    5: "objects_in_area",
    6: "destroy_object",
    7: "capture_object",
    8: "accumulate_attribute",
    9: "research_technology",
    10: "timer",
    11: "object_selected",
    12: "ai_signal",
    13: "player_defeated",
    14: "object_has_target",
    15: "object_visible",
    16: "object_not_visible",
    17: "researching_tech",
    18: "units_garrisoned",
    19: "difficulty_level",
    20: "chance",
    21: "technology_state",
    22: "variable_value",
    23: "object_hp",
    24: "diplomacy_state",
})

empty_attributes = {
    "condition_type": -1,
    "amount_or_quantity": -1,
    "resource_type_or_tribute_list": -1,
    "unit_object": -1,
    "next_object": -1,
    "object_list": -1,
    "player": -1,
    "technology": -1,
    "timer": -1,
    "area_1_x": -1,
    "area_1_y": -1,
    "area_2_x": -1,
    "area_2_y": -1,
    "object_group": -1,
    "object_type": -1,
    "ai_signal": -1,
    "inverted": -1,
    "variable": -1,
    "comparison": -1,
    "target_player": -1,
}

default_attributes = {
    0: {
        "condition_type": 0,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": -1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": -1,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    8: {
        "condition_type": 8,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": 0,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    12: {
        "condition_type": 12,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": -1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": 0,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    1: {
        "condition_type": 1,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": -1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": 0,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    2: {
        "condition_type": 2,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": -1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    7: {
        "condition_type": 7,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    20: {
        "condition_type": 20,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    6: {
        "condition_type": 6,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    19: {
        "condition_type": 19,
        "amount_or_quantity": 3,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": -1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    24: {
        "condition_type": 24,
        "amount_or_quantity": 0,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": 1,
    },
    14: {
        "condition_type": 14,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": 1,
    },
    23: {
        "condition_type": 23,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": 0,
        "target_player": 1,
    },
    16: {
        "condition_type": 16,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": 0,
        "target_player": 1,
    },
    11: {
        "condition_type": 11,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": 0,
        "target_player": 1,
    },
    15: {
        "condition_type": 15,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": 0,
        "target_player": 1,
    },
    5: {
        "condition_type": 5,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": 0,
        "target_player": 1,
    },
    4: {
        "condition_type": 4,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": 0,
        "target_player": 1,
    },
    3: {
        "condition_type": 3,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": 0,
        "target_player": 1,
    },
    13: {
        "condition_type": 13,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": -1,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    9: {
        "condition_type": 9,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": 16,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    17: {
        "condition_type": 17,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": 16,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    21: {
        "condition_type": 21,
        "amount_or_quantity": 3,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": 1,
        "technology": 16,
        "timer": -1,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    10: {
        "condition_type": 10,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": -1,
        "technology": -1,
        "timer": 10,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    18: {
        "condition_type": 18,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": -1,
        "technology": -1,
        "timer": 10,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": -1,
        "comparison": -1,
        "target_player": -1,
    },
    22: {
        "condition_type": 22,
        "amount_or_quantity": -1,
        "resource_type_or_tribute_list": -1,
        "unit_object": -1,
        "next_object": -1,
        "object_list": -1,
        "player": -1,
        "technology": -1,
        "timer": 10,
        "area_1_x": -1,
        "area_1_y": -1,
        "area_2_x": -1,
        "area_2_y": -1,
        "object_group": -1,
        "object_type": -1,
        "ai_signal": -1,
        "inverted": 0,
        "variable": 0,
        "comparison": 0,
        "target_player": -1,
    }
}

attributes = {
    0: [
        "condition_type",
    ],
    1: [
        "condition_type",
        "unit_object",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "inverted",
    ],
    2: [
        "condition_type",
        "unit_object",
        "next_object",
        "inverted",
    ],
    3: [
        "condition_type",
        "amount_or_quantity",
        "object_list",
        "player",
        "object_group",
        "object_type",
    ],
    4: [
        "condition_type",
        "amount_or_quantity",
        "object_list",
        "player",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
    ],
    5: [
        "condition_type",
        "amount_or_quantity",
        "object_list",
        "player",
        "area_1_x",
        "area_1_y",
        "area_2_x",
        "area_2_y",
        "object_group",
        "object_type",
        "inverted",
    ],
    6: [
        "condition_type",
        "unit_object",
        "inverted",
    ],
    7: [
        "condition_type",
        "unit_object",
        "player",
        "inverted",
    ],
    8: [
        "condition_type",
        "amount_or_quantity",
        "resource_type_or_tribute_list",
        "player",
        "inverted",
    ],
    9: [
        "condition_type",
        "player",
        "technology",
        "inverted",
    ],
    10: [
        "condition_type",
        "timer",
        "inverted",
    ],
    11: [
        "condition_type",
        "unit_object",
        "inverted",
    ],
    12: [
        "condition_type",
        "ai_signal",
        "inverted",
    ],
    13: [
        "condition_type",
        "player",
        "inverted",
    ],
    14: [
        "condition_type",
        "unit_object",
        "next_object",
        "object_list",
        "object_group",
        "object_type",
        "inverted",
    ],
    15: [
        "condition_type",
        "unit_object",
    ],
    16: [
        "condition_type",
        "unit_object",
    ],
    17: [
        "condition_type",
        "player",
        "technology",
        "inverted",
    ],
    18: [
        "condition_type",
        "amount_or_quantity",
        "unit_object",
        "inverted",
    ],
    19: [
        "condition_type",
        "amount_or_quantity",
        "inverted",
    ],
    20: [
        "condition_type",
        "amount_or_quantity",
    ],
    21: [
        "condition_type",
        "amount_or_quantity",
        "player",
        "technology",
        "inverted",
    ],
    22: [
        "condition_type",
        "amount_or_quantity",
        "inverted",
        "variable",
        "comparison",
    ],
    23: [
        "condition_type",
        "amount_or_quantity",
        "unit_object",
        "inverted",
        "comparison",
    ],
    24: [
        "condition_type",
        "amount_or_quantity",
        "player",
        "inverted",
        "target_player",
    ]
}

# for condition_id in range(0, 25):
#     try:
#         params = parameters[condition_id]
#         print("\"\"\"Attributes for the **xxxxx** condition are:\\n", end="")
#         for parameter in params:
#             if parameter is not "condition_type":
#                 print("\n-", parameter, end="")
#     except KeyError:
#         continue
#     print("\"\"\"")



from AoE2ScenarioParser.helper import helper
from AoE2ScenarioParser.helper.datatype import DataType


class Retriever:
    """ A Class for defining how to retrieve data.
    The Constructor has quite some parameters which can all be used for getting the proper data
        name: The name of the item. Recommended to make this unique within the Piece or Struct
        datatype: A datatype object
        on_success: A lambda which can be used to execute some code once the data is retrieved. You can use one
                    parameter; the data that got retrieved.
        save_as: Save the value retrieved to a dict. Can be used later with the 'set_repeat' parameter
        set_repeat: Use a saved value (from 'save_as') to set repeat value. You can use values from the dict by
                    surrounding them with curly brackets. The string will be executed using eval.
        log_value: A boolean for, mostly, debugging. This will log this Retriever with it's data on retrieval.
    """

    def __init__(self, name, datatype=DataType(), on_success=None, save_as=None, set_repeat=None, log_value=False):
        self.name = name
        self.datatype = datatype
        self.on_success = on_success
        self.save_as = save_as
        self.set_repeat = set_repeat
        self.log_value = log_value
        self.data = None

    def set_data(self, data):
        self.data = data
        if self.on_success is not None:
            self.on_success(data)

    def get_short_str(self):
        if self.data is not None:
            if type(self.data) is str:
                return self.name + " (" + self.datatype.to_simple_string() + "): '" + str(self.data) + "'"
            else:
                return self.name + " (" + self.datatype.to_simple_string() + "): " + str(self.data) + ""

    def __repr__(self):
        if type(self.data) is list:
            data = str(helper.pretty_print_list(self.data))
        else:
            data = str(self.data)
        return "[Retriever] " + self.name + ": " + str(self.datatype) + " >>> " + str(data)


def find_retriever(retriever_list, name):
    for x in retriever_list:
        if x.name == name:
            return x



from __future__ import annotations

from enum import IntEnum


class Player(IntEnum):
    GAIA = 0
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    EIGHT = 8


class PlayerColor(IntEnum):
    BLUE = 1
    RED = 2
    GREEN = 3
    YELLOW = 4
    AQUA = 5
    PURPLE = 6
    GREY = 7
    ORANGE = 8



from AoE2ScenarioParser.helper.datatype import DataType
from AoE2ScenarioParser.helper.retriever import Retriever
from AoE2ScenarioParser.pieces import aoe2_piece
from AoE2ScenarioParser.pieces.structs.trigger import TriggerStruct
from AoE2ScenarioParser.pieces.structs.variable import VariableStruct


class TriggerPiece(aoe2_piece.AoE2Piece):
    def __init__(self, parser_obj=None, data=None):
        retrievers = [
            Retriever("Trigger Version", DataType("f64")),
            Retriever("Trigger instructions start", DataType("s8")),
            Retriever("Number of triggers", DataType("s32"), save_as="number_of_triggers"),
            Retriever("Trigger data", DataType(TriggerStruct), set_repeat="{number_of_triggers}"),
            Retriever("Trigger display order array", DataType("u32"), set_repeat="{number_of_triggers}"),
            Retriever("Unknown Data", DataType("1028")),
            Retriever("Number of variables", DataType("u32"), save_as="number_of_vars"),
            Retriever("Variables", DataType(VariableStruct), set_repeat="{number_of_vars}")
        ]

        super().__init__("Triggers", retrievers, parser_obj, data=data)



import pyautogui
from time import sleep

for x in range(5, 0, -1):
    print(x, end=" ")
    sleep(1)
print("\n")

# Change to 'tabbing mode'
pyautogui.press('tab')
sleep(0.2)

for _ in range(0, 9):
    pyautogui.press('tab')

while True:
    pyautogui.hotkey('shiftleft', 'tab')
    pyautogui.hotkey('shiftleft', 'tab')
    pyautogui.hotkey('shiftleft', 'tab')
    pyautogui.hotkey('shiftleft', 'tab')
    sleep(0.05)
    pyautogui.press('enter')
    sleep(0.05)
    pyautogui.press('tab')
    pyautogui.press('tab')
    pyautogui.press('tab')
    pyautogui.press('tab')
    pyautogui.press('enter')
    pyautogui.press('down')
    pyautogui.press('enter')
